#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import sys
from typing import List, Optional, TextIO, Tuple, cast

import click

from aws_orbit.commands import deploy as deploy_commands
from aws_orbit.commands import destroy as destroy_commands
from aws_orbit.commands.build import build_image, build_podsetting
from aws_orbit.commands.delete import delete_image, delete_podsetting
from aws_orbit.commands.init import init
from aws_orbit.commands.list import list_env, list_images
from aws_orbit.utils import print_dir

DEBUG_LOGGING_FORMAT = "[%(asctime)s][%(filename)-13s:%(lineno)3d] %(message)s"
DEBUG_LOGGING_FORMAT_REMOTE = "[%(filename)-13s:%(lineno)3d] %(message)s"
_logger: logging.Logger = logging.getLogger(__name__)


def enable_debug(format: str) -> None:
    logging.basicConfig(level=logging.DEBUG, format=format)
    _logger.setLevel(logging.DEBUG)
    logging.getLogger("boto3").setLevel(logging.ERROR)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("s3transfer").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)
    logging.getLogger("sh").setLevel(logging.ERROR)
    logging.getLogger("kubernetes").setLevel(logging.ERROR)


@click.group()
def cli() -> None:
    """Orbit Workbench CLI - Data & ML Unified Development and Production Environment"""
    pass


@click.command(name="init")
@click.option(
    "--name",
    "-n",
    type=str,
    help="The name of the Orbit Workbench enviroment. MUST be unique per AWS account.",
    required=False,
    default="my-env",
    show_default=True,
)
@click.option(
    "--region",
    "-r",
    type=str,
    default=None,
    help="AWS Region name (e.g. us-east-1). If None, it will be infered.",
    show_default=False,
    required=False,
)
@click.option(
    "--foundation/--no-foundation", default=True, help="Create Orbit foundation default manifest.", show_default=True
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def init_cli(
    name: str,
    region: Optional[str],
    foundation: bool,
    debug: bool,
) -> None:
    """Creates a Orbit Workbench manifest model file (yaml) where all your deployment settings will rest."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("name: %s", name)
    _logger.debug("region: %s", region)
    _logger.debug("foundation: %s", foundation)
    _logger.debug("debug: %s", debug)
    init(name=name, region=region, foundation=foundation, debug=debug)


@click.group(name="deploy")
def deploy() -> None:
    """Deploy foundation,env,teams in your Orbit Workbench."""
    pass


@deploy.command(name="toolkit")
@click.option(
    "--filename",
    "-f",
    type=str,
    help="The target Orbit Workbench manifest file (yaml).",
    required=True,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_toolkit(
    filename: str,
    debug: bool,
) -> None:
    """Deploy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    filename = filename if filename[0] in (".", "/") else f"./{filename}"
    _logger.debug("filename: %s", filename)
    deploy_commands.deploy_toolkit(
        filename=filename,
        debug=debug,
    )


@deploy.command(name="credentials")
@click.option(
    "--filename",
    "-f",
    type=str,
    help="The target Orbit Workbench manifest file (yaml).",
)
@click.option("--username", "-u", type=str, help="Image Registry username", required=True)
@click.option(
    "--password",
    "-p",
    type=str,
    help="Image Registry password",
)
@click.option("--registry", "-r", type=str, help="Image Registry name/URL", default="docker.io")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_credentials(
    filename: str,
    username: str,
    password: str,
    registry: str,
    debug: bool,
) -> None:
    """Deploy Image Registry credentials for use in building and pulling images"""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("filename: %s", filename)
    _logger.debug("username: %s", username)
    _logger.debug("registry: %s", registry)
    deploy_commands.deploy_credentials(
        filename=filename,
        username=username,
        password=password,
        registry=registry,
        debug=debug,
    )


@deploy.command(name="teams")
@click.option(
    "--filename",
    "-f",
    type=str,
    help="The target Orbit Workbench manifest file (yaml).",
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_teams(
    filename: str,
    debug: bool,
) -> None:
    """Deploy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    filename = filename if filename[0] in (".", "/") else f"./{filename}"
    _logger.debug("filename: %s", filename)
    deploy_commands.deploy_teams(
        filename=filename,
        debug=debug,
    )


@deploy.command(name="env")
@click.option(
    "--filename",
    "-f",
    type=str,
    help="The target Orbit Workbench manifest file (yaml).",
    required=True,
)
@click.option(
    "--skip-images/--build-images",
    default=True,
    help="Skip Docker images updates (Usually for development purpose).",
    show_default=True,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_env(
    filename: str,
    skip_images: bool,
    debug: bool,
) -> None:
    """Deploy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    filename = filename if filename[0] in (".", "/") else f"./{filename}"
    _logger.debug("filename: %s", filename)
    _logger.debug("skip_images: %s", skip_images)
    deploy_commands.deploy_env(
        filename=filename,
        skip_images=skip_images,
        debug=debug,
    )


@deploy.command(name="foundation")
@click.option(
    "--filename",
    "-f",
    type=str,
    help="The target Orbit Workbench manifest file (yaml).",
)
@click.option(
    "--name",
    "-n",
    type=str,
    help="The Name of the Orbit Foundation deployment",
)
@click.option(
    "--codeartifact-domain",
    type=str,
    help="CodeArtifact Domain to pull packages from.",
)
@click.option(
    "--codeartifact-repository",
    type=str,
    help="CodeArtifact Repository to pull packages from.",
)
@click.option(
    "--ssl-cert-arn",
    type=str,
    help="SSL Certificate to integrate the ALB with.",
)
@click.option(
    "--custom-domain-name",
    type=str,
    help="The Custom Domain Name to associate the orbit framework with",
)
@click.option(
    "--role-prefix",
    type=str,
    help="The Prefix to attach to the IAM Role",
)
@click.option(
    "--internet-accessibility/--no-internet-accessibility",
    default=True,
    help="Configure for deployment to Private (internet accessibility) "
    "or Isolated (no internet accessibility) subnets.",
    show_default=True,
)
@click.option(
    "--max-availability-zones",
    default=2,
    help="The maximum number of Availability Zones to attempt to deploy in the VPC",
    show_default=True,
    type=int,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_foundation(
    filename: Optional[str] = None,
    name: Optional[str] = None,
    debug: bool = False,
    internet_accessibility: bool = True,
    codeartifact_domain: Optional[str] = None,
    codeartifact_repository: Optional[str] = None,
    ssl_cert_arn: Optional[str] = None,
    custom_domain_name: Optional[str] = None,
    max_availability_zones: Optional[int] = None,
    role_prefix: Optional[str] = None,
) -> None:
    """Deploy a Orbit Workbench foundation based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    if not filename and not name:
        raise click.ClickException('One of "filename" or "name" is required.')

    if filename:
        filename = filename if filename[0] in (".", "/") else f"./{filename}"
    _logger.debug("filename: %s", filename)
    _logger.debug("name: %s", name)
    _logger.debug("codeartifact_domain: %s", codeartifact_domain)
    _logger.debug("codeartifact_repository: %s", codeartifact_repository)
    _logger.debug("ssl_cert_arn: %s", ssl_cert_arn)
    _logger.debug("custom_domain_name: %s", custom_domain_name)
    _logger.debug("max_availability_zones: %s", max_availability_zones)
    _logger.debug("role_prefix: %s", role_prefix)
    deploy_commands.deploy_foundation(
        filename=filename,
        name=name,
        codeartifact_domain=codeartifact_domain,
        codeartifact_repository=codeartifact_repository,
        ssl_cert_arn=ssl_cert_arn,
        custom_domain_name=custom_domain_name,
        debug=debug,
        internet_accessibility=internet_accessibility,
        max_availability_zones=max_availability_zones,
        role_prefix=role_prefix,
    )


@click.group(name="destroy")
def destroy() -> None:
    """Destroy foundation,env,etc in your Orbit Workbench."""
    pass


@destroy.command(name="teams")
@click.option("--env", "-e", type=str, required=True, help="Destroy Orbit Teams.")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def destroy_teams(env: str, debug: bool) -> None:
    """Destroy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    destroy_commands.destroy_teams(env=env, debug=debug)


@destroy.command(name="env")
@click.option("--env", "-e", type=str, required=True, help="Destroy Orbit Environment.")
@click.option(
    "--preserve-credentials/--destroy-credentials",
    default=True,
    help="Preserve any docker credentials in Secrets Manager.",
    show_default=True,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def destroy_env(env: str, preserve_credentials: bool, debug: bool) -> None:
    """Destroy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    destroy_commands.destroy_env(env=env, preserve_credentials=preserve_credentials, debug=debug)


@destroy.command(name="foundation")
@click.option("--name", "-n", type=str, required=True, help="Destroy Orbit Foundation.")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def destroy_foundation(name: str, debug: bool) -> None:
    """Destroy a Orbit Workbench environment based on a manisfest file (yaml)."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("name: %s", name)
    destroy_commands.destroy_foundation(env=name, debug=debug)


@destroy.command(name="credentials")
@click.option("--env", "-e", type=str, required=True, help="Destroy Registry Credentials.")
@click.option("--registry", "-r", type=str, required=True, help="Image Registry.")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def destroy_credentials(env: str, registry: str, debug: bool) -> None:
    """Destroy Image Registry Credentials previously stored"""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    _logger.debug("registry: %s", registry)
    destroy_commands.destroy_credentials(env=env, registry=registry, debug=debug)


@click.group(name="build")
def build() -> None:
    """Build images,podsettings,etc in your Orbit Workbench."""
    pass


@build.command(name="image")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment.")
@click.option("--dir", "-d", type=str, help="Dockerfile directory.", required=True)
@click.option("--name", "-n", type=str, help="Image name.", required=True)
@click.option("--timeout", type=int, help="CodeBuild Timeout", default=30, show_default=True)
@click.option(
    "--script",
    "-s",
    type=str,
    default=None,
    help="Build script to run before the image build.",
    required=False,
)
@click.option(
    "--build-arg",
    type=str,
    multiple=True,
    default=[],
    help="One or more --build-arg parameters to pass to the Docker build command.",
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def deploy_image_cli(
    env: str,
    dir: str,
    name: str,
    timeout: Optional[int],
    script: Optional[str],
    build_arg: Optional[List[str]],
    debug: bool,
) -> None:
    """Build and Deploy a new Docker image into ECR."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    _logger.debug("dir: %s", dir)
    _logger.debug("name: %s", name)
    _logger.debug("script: %s", script)
    _logger.debug("timeout: %s", timeout)
    _logger.debug("debug: %s", debug)
    build_image(
        dir=dir, name=name, env=env, timeout=cast(int, timeout), script=script, build_args=build_arg, debug=debug
    )


@click.group(name="replicate")
def replicate() -> None:
    """Replicate images from external respositories into ECR"""
    pass


@replicate.command(name="image")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment.")
@click.option("--name", "-n", type=str, help="Image name.", required=True)
@click.option(
    "--script",
    "-s",
    type=str,
    default=None,
    help="Build script to run before the image build.",
    required=False,
)
@click.option(
    "--source-registry",
    "-y",
    type=str,
    multiple=False,
    default=None,
    help="One or more Teams to deploy the image to (can de declared multiple times).",
    required=True,
)
@click.option(
    "--source-repository",
    "-r",
    type=str,
    multiple=False,
    default=None,
    help="One or more Teams to deploy the image to (can de declared multiple times).",
    required=True,
)
@click.option(
    "--source-version",
    "-v",
    type=str,
    multiple=False,
    default=None,
    help="One or more Teams to deploy the image to (can de declared multiple times).",
    required=True,
)
@click.option(
    "--build-arg",
    type=str,
    multiple=True,
    default=[],
    help="One or more --build-arg parameters to pass to the Docker build command.",
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def replicate_image_cli(
    env: str,
    name: str,
    script: Optional[str],
    source_registry: str,
    source_repository: str,
    source_version: str,
    build_arg: Optional[List[str]],
    debug: bool,
) -> None:
    """Build and Deploy a new Docker image into ECR."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    _logger.debug("name: %s", name)
    _logger.debug("script: %s", script)
    _logger.debug("debug: %s", debug)
    build_image(
        dir=None,
        name=name,
        env=env,
        script=script,
        source_registry=source_registry,
        source_repository=source_repository,
        source_version=source_version,
        build_args=build_arg,
        debug=debug,
    )


@build.command(name="profile")
@click.option("--env", "-e", type=str, required=False, help="Orbit Environment.")
@click.option("--team", "-t", type=str, help="Orbit Team.", required=False)
@click.argument("profile", type=click.File("r"), required=False)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def add_profile_cli(env: str, team: str, debug: bool, profile: TextIO) -> None:
    raise click.ClickException("Profiles are no longer supported...please use podsettings")


@build.command(name="podsetting")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment.")
@click.option("--team", "-t", type=str, required=True, help="Orbit Team.")
@click.argument("podsetting", type=click.File("r"))
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def add_podsetting_cli(env: str, team: str, debug: bool, podsetting: TextIO) -> None:
    """Deploy a new podsetting to the K8s cluster."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    _logger.debug("team: %s", team)
    podsetting_str = podsetting.read()
    _logger.debug("podsetting: %s", podsetting_str)
    _logger.debug("debug: %s", debug)
    try:
        build_podsetting(env_name=env, team_name=team, podsetting=podsetting_str, debug=debug)
    except ImportError:
        raise click.ClickException('The "utils" submodule is required to use "run" commands')


@click.group(name="delete")
def delete() -> None:
    """Delete images,podsettings,etc in your Orbit Workbench."""
    pass


@delete.command(name="profile")
@click.option("--env", "-e", type=str, required=False, help="Orbit Environment.")
@click.option("--team", "-t", type=str, help="Orbit Team.", required=False)
@click.option("--profile", "-p", type=str, help="Profile name to delete", required=False)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def delete_profile_cli(env: str, team: str, profile: str, debug: bool) -> None:
    raise click.ClickException("Profiles are no longer supported...please use podsettings")


@delete.command(name="podsetting")
@click.option("--team", "-t", type=str, help="Orbit Team.", required=True)
@click.option("--podsetting", "-n", type=str, help="Podsetting name to delete", required=True)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def delete_podsetting_cli(team: str, podsetting: str, debug: bool) -> None:
    """Delete a podsetting"""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("team: %s", team)
    _logger.debug("podsetting: %s", podsetting)
    _logger.debug("debug: %s", debug)
    delete_podsetting(namespace=team, podsetting_name=podsetting, debug=debug)


@delete.command(name="image")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment.")
@click.option("--name", "-n", type=str, help="Image name.", required=True)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def delete_image_cli(env: str, name: str, debug: bool) -> None:
    """Destroy a Docker image from ECR."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("env: %s", env)
    _logger.debug("name: %s", name)
    _logger.debug("debug: %s", debug)
    delete_image(name=name, env=env, debug=debug)


@click.group(name="list")
def list() -> None:
    """List images, etc in your Orbit Workbench."""
    pass


@list.command(name="profile")
@click.option("--env", "-e", type=str, required=False, help="Orbit Environment.")
@click.option("--team", "-t", type=str, help="Orbit Team.", required=False)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def list_profiles_cli(env: str, team: str, debug: bool) -> None:
    raise click.ClickException("Profiles are no longer supported.")


@list.command(name="image")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment.")
@click.option(
    "--region",
    "-r",
    type=str,
    default=None,
    help="AWS Region name (e.g. us-east-1). If None, it will be infered.",
    show_default=False,
    required=False,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
def list_images_cli(env: str, region: Optional[str], debug: bool) -> None:
    """List all Docker images available into the target environment."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    list_images(env=env, region=region)


@list.command(name="env")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@click.option("--env", "-e", type=str, required=False, default="", help="Select a single Orbit environment")
@click.option(
    "--variable",
    type=click.Choice(["all", "landing-page", "teams", "toolkitbucket"], case_sensitive=False),
    show_default=True,
    default="all",
)
def list_env_cli(debug: bool, env: str, variable: str) -> None:
    """List all Docker images available into the target environment."""
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    list_env(env, variable)


@click.command(name="remote", hidden=True)
@click.option("--command", "-c", type=str, required=True)
@click.argument("args", nargs=-1)
def remote_cli(command: str, args: Tuple[str]) -> None:
    """Run command remotely on CodeBuild"""
    enable_debug(format=DEBUG_LOGGING_FORMAT_REMOTE)
    from aws_orbit.remote_files import REMOTE_FUNC_TYPE, RemoteCommands

    _logger.debug("Remote bundle structure:")
    print_dir(os.getcwd(), exclude=["__pycache__", "cdk", ".venv", ".mypy_cache"])
    remote_func: REMOTE_FUNC_TYPE = getattr(RemoteCommands, command)
    remote_func(args)


@click.group(name="run")
def run_container() -> None:
    """Execute containers in the Orbit environment"""
    try:
        import aws_orbit_sdk  # noqa: F401
    except ImportError:
        raise click.ClickException('The "utils" submodule is required to use "run" commands')
    pass


@run_container.command(name="python", help="Run python script in a container")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment to execute container in.")
@click.option("--team", "-t", type=str, required=True, help="Orbit Team Space to execute container in.")
@click.option(
    "--user", "-u", type=str, default="jovyan", show_default=True, help="Jupyter user to execute container as."
)
@click.option("--wait/--no-wait", type=bool, default=False, show_default=True, help="Wait for execution to complete.")
@click.option(
    "--delay",
    type=int,
    required=False,
    help="If --wait, this is the number of seconds to sleep between container state checks.",
)
@click.option(
    "--max-attempts",
    type=int,
    required=False,
    help="If --wait, this is the number of times to check container state before failing.",
)
@click.option(
    "--tail-logs/--no-tail-logs",
    type=bool,
    default=False,
    show_default=True,
    help="If --wait, print a tail of container logs after execution completes.",
)
@click.option(
    "--debug/--no-debug",
    default=False,
    show_default=True,
    help="Enable detailed logging.",
)
@click.argument("input", type=click.File("r"))
def run_python_container(
    env: str,
    team: str,
    user: str,
    wait: bool,
    delay: Optional[int],
    max_attempts: Optional[int],
    tail_logs: bool,
    debug: bool,
    input: TextIO,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    _logger.debug("env: %s", env)
    _logger.debug("team: %s", team)
    _logger.debug("user: %s", user)

    tasks = json.load(input)
    _logger.debug("tasks: %s", json.dumps(tasks))

    import aws_orbit.commands.run as run

    no_error_flag = run.run_python_container(
        env=env,
        team=team,
        user=user,
        tasks=tasks,
        wait=wait,
        delay=delay,
        max_attempts=max_attempts,
        tail_logs=tail_logs,
        debug=debug,
    )
    if no_error_flag:
        sys.exit(0)
    else:
        sys.exit(1)


@run_container.command(name="notebook", help="Run notebook in a container")
@click.option("--env", "-e", type=str, required=True, help="Orbit Environment to execute container in.")
@click.option("--team", "-t", type=str, required=True, help="Orbit Team Space to execute container in.")
@click.option(
    "--user", "-u", type=str, default="jovyan", show_default=True, help="Jupyter user to execute container as."
)
@click.option("--wait/--no-wait", type=bool, default=False, show_default=True, help="Wait for execution to complete.")
@click.option(
    "--delay",
    type=int,
    required=False,
    help="If --wait, this is the number of seconds to sleep between container state checks.",
)
@click.option(
    "--max-attempts",
    type=int,
    required=False,
    help="If --wait, this is the number of times to check container state before failing.",
)
@click.option(
    "--tail-logs/--no-tail-logs",
    type=bool,
    default=False,
    show_default=True,
    help="If --wait, print a tail of container logs after execution completes.",
)
@click.option(
    "--debug/--no-debug",
    default=False,
    show_default=True,
    help="Enable detailed logging.",
)
@click.argument("input", type=click.File("r"))
def run_notebook_container(
    env: str,
    team: str,
    user: str,
    wait: bool,
    delay: Optional[int],
    max_attempts: Optional[int],
    tail_logs: bool,
    debug: bool,
    input: TextIO,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    _logger.debug("env: %s", env)
    _logger.debug("team: %s", team)
    _logger.debug("user: %s", user)

    tasks = json.load(input)
    _logger.debug("tasks: %s", json.dumps(tasks))

    import aws_orbit.commands.run as run

    no_error_flag = run.run_notebook_container(
        env=env,
        team=team,
        user=user,
        tasks=tasks,
        wait=wait,
        delay=delay,
        max_attempts=max_attempts,
        tail_logs=tail_logs,
        debug=debug,
    )
    if no_error_flag:
        sys.exit(0)
    else:
        sys.exit(1)


def main() -> int:
    # For now , we will not support init command and the deploy page points to manifest examples to be used.
    # cli.add_command(init_cli)
    cli.add_command(deploy)
    cli.add_command(destroy)
    cli.add_command(remote_cli)
    cli.add_command(run_container)
    cli.add_command(build)
    cli.add_command(delete)
    cli.add_command(list)
    cli.add_command(replicate)
    cli()
    return 0



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from pprint import pformat
from typing import Any, Dict, List

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _tags2dict(tags: List[Dict[str, str]]) -> Dict[str, str]:
    return {x["Key"]: x["Value"] for x in tags}


def _check_elb_tag(env_name: str, tags: Dict[str, str]) -> bool:
    key: str = f"kubernetes.io/cluster/orbit-{env_name}"
    return tags.get(key) == "owned"


def _apply_tag_filter(env_name: str, elbs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    names: List[str] = [x["LoadBalancerName"] for x in elbs]
    _logger.debug("Filtering ELBS:\n%s", pformat(names))
    if not names:
        return []
    client = boto3_client("elb")
    res: List[Dict[str, Any]] = client.describe_tags(LoadBalancerNames=names)["TagDescriptions"]
    tags: Dict[str, Dict[str, str]] = {x["LoadBalancerName"]: _tags2dict(tags=x["Tags"]) for x in res}
    return [x for x in elbs if _check_elb_tag(env_name=env_name, tags=tags[x["LoadBalancerName"]])]


def describe_load_balancers(env_name: str) -> List[Dict[str, Any]]:
    paginator = boto3_client("elb").get_paginator("describe_load_balancers")
    elbs: List[Dict[str, Any]] = []
    for page in paginator.paginate():
        _logger.debug(page)
        elbs += _apply_tag_filter(env_name=env_name, elbs=page.get("LoadBalancerDescriptions", []))
    _logger.debug("Filtered ELB names:\n%s", pformat([x["LoadBalancerName"] for x in elbs]))
    return elbs


def identify_services(names: List[str]) -> Dict[str, str]:
    if not names:
        return {}
    client = boto3_client("elb")
    res: List[Dict[str, Any]] = client.describe_tags(LoadBalancerNames=names)["TagDescriptions"]
    tags: Dict[str, Dict[str, str]] = {x["LoadBalancerName"]: _tags2dict(tags=x["Tags"]) for x in res}
    services = {t["kubernetes.io/service-name"]: name for name, t in tags.items() if "kubernetes.io/service-name" in t}
    _logger.debug("Services itentified for each ELB:\n%s", pformat(services))
    return services


def _search_elb_by_name(elbs: List[Dict[str, Any]], name: str) -> Dict[str, Any]:
    for elb in elbs:
        if elb["LoadBalancerName"] == name:
            return elb
    raise RuntimeError(f"ELB {name} not found!")


def get_elbs_by_service(env_name: str) -> Dict[str, Dict[str, Any]]:
    # Cleaning up CreatedTime cause datatime objects are not JSON serializable
    elbs = [
        {
            k: v
            for k, v in x.items()
            if k
            not in {
                "CreatedTime",
                "HealthCheck",
                "CanonicalHostedZoneNameID",
                "CanonicalHostedZoneName",
                "BackendServerDescriptions",
            }
        }
        for x in describe_load_balancers(env_name=env_name)
    ]
    services = identify_services(names=[x["LoadBalancerName"] for x in elbs])
    return {s: _search_elb_by_name(elbs=elbs, name=a) for s, a in services.items()}


def delete_load_balancers(env_name: str) -> None:
    client = boto3_client("elb")
    for elb in describe_load_balancers(env_name=env_name):
        client.delete_load_balancer(LoadBalancerName=elb["LoadBalancerName"])



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import time
from typing import Any, Dict

from aws_orbit import ORBIT_CLI_ROOT, sh, utils
from aws_orbit.exceptions import FailedShellCommand
from aws_orbit.models.context import Context
from aws_orbit.remote_files import kubectl
from aws_orbit.remote_files.utils import get_k8s_context
from aws_orbit.services import cfn
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)

CONFIG_PATH = os.path.join(ORBIT_CLI_ROOT, "data", "kubeflow")


def _cleanup_output(output_path: str) -> None:
    files = os.listdir(output_path)
    for file in files:
        if file.endswith(".yaml"):
            os.remove(os.path.join(output_path, file))


def gen_kubeflow_config(context: Context, output_path: str, cluster_name: str) -> None:

    os.makedirs(output_path, exist_ok=True)
    _cleanup_output(output_path=output_path)
    if context.account_id is None:
        raise RuntimeError("context.account_id is None!")
    if context.region is None:
        raise RuntimeError("context.region is None!")

    input = os.path.join(CONFIG_PATH, "kfctl_aws.yaml")
    output = os.path.join(output_path, "kfctl_aws.yaml")

    client = boto3_client(service_name="cognito-idp")
    response: Dict[str, Any] = client.describe_user_pool(UserPoolId=context.user_pool_id)
    domain: str = response["UserPool"].get("Domain")

    with open(input, "r") as file:
        content: str = file.read()

    content = utils.resolve_parameters(
        content,
        dict(
            certArn=context.networking.frontend.ssl_cert_arn,
            cognitoAppClientId=context.user_pool_client_id,
            cognitoUserPoolID=context.user_pool_id,
            account_id=context.account_id,
            region=context.region,
            env_name=context.name,
            cluster_name=cluster_name,
            cognitoUserPoolDomain=domain,
        ),
    )
    _logger.debug("Kubeflow configuration:\n%s", content)
    with open(output, "w") as file:
        file.write(content)

    k8s_context = get_k8s_context(context=context)

    input = os.path.join(CONFIG_PATH, "apply_kf.sh")
    output = os.path.join(output_path, "apply_kf.sh")

    with open(input, "r") as file:
        content = file.read()

    content = utils.resolve_parameters(
        content,
        dict(cluster_name=cluster_name, k8s_context=k8s_context),
    )
    _logger.debug("Kubeflow script:\n%s", content)
    with open(output, "w") as file:
        file.write(content)

    sh.run(f"chmod a+x  {output}")

    input = os.path.join(CONFIG_PATH, "delete_kf.sh")
    output = os.path.join(output_path, "delete_kf.sh")

    with open(input, "r") as file:
        content = file.read()

    content = utils.resolve_parameters(
        content,
        dict(cluster_name=cluster_name, k8s_context=k8s_context),
    )
    _logger.debug("Kubeflow script:\n%s", content)
    with open(output, "w") as file:
        file.write(content)

    sh.run(f"chmod a+x  {output}")


def deploy_kubeflow(context: Context) -> None:
    stack_name: str = f"orbit-{context.name}"
    final_eks_stack_name: str = f"eksctl-{stack_name}-cluster"
    _logger.debug("EKSCTL stack name: %s", final_eks_stack_name)

    if cfn.does_stack_exist(stack_name=final_eks_stack_name):
        cluster_name = f"orbit-{context.name}"

        output_path = os.path.join(".orbit.out", context.name, "kubeflow", cluster_name)
        gen_kubeflow_config(context, output_path, cluster_name)

        _logger.debug("Deploying Kubeflow")
        output_path = os.path.abspath(output_path)
        _logger.debug(f"kubeflow config dir: {output_path}")
        utils.print_dir(output_path)
        sh.run("./apply_kf.sh", cwd=output_path)
        time.sleep(120)
        sh.run("kubectl delete ds -n kubeflow nvidia-device-plugin-daemonset")


def destroy_kubeflow(context: Context) -> None:
    stack_name: str = f"orbit-{context.name}"
    final_eks_stack_name: str = f"eksctl-{stack_name}-cluster"
    _logger.debug("EKSCTL stack name: %s", final_eks_stack_name)

    if cfn.does_stack_exist(stack_name=final_eks_stack_name):
        kubectl.write_kubeconfig(context=context)

        for line in sh.run_iterating("kubectl get namespace kubeflow"):
            if '"kubeflow" not found' in line:
                return

        cluster_name = f"orbit-{context.name}"
        output_path = os.path.join(".orbit.out", context.name, "kubeflow", cluster_name)
        gen_kubeflow_config(context, output_path, cluster_name)

        _logger.debug("Destroying Kubeflow")
        output_path = os.path.abspath(output_path)
        _logger.debug(f"kubeflow config dir: {output_path}")
        utils.print_dir(output_path)

        timeouts = 0
        while timeouts < 3:
            try:
                _logger.info("Deleting kubeflow resources")
                sh.run("./delete_kf.sh", cwd=output_path)
            except FailedShellCommand:
                _logger.info("The command returned a non-zero exit code. Retrying to delete resources")
                timeouts += 1
                time.sleep(300)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import ipaddress
import logging
import math
import os
import random
import time
from string import Template
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Union

import boto3
import botocore.exceptions

from aws_orbit import __version__

if TYPE_CHECKING:
    from aws_orbit.models.context import Context

_logger: logging.Logger = logging.getLogger(__name__)


def chunkify(lst: List[Any], num_chunks: int = 1, max_length: Optional[int] = None) -> List[List[Any]]:
    num: int = num_chunks if max_length is None else int(math.ceil((float(len(lst)) / float(max_length))))
    return [lst[i : i + num] for i in range(0, len(lst), num)]  # noqa: E203


def get_botocore_config() -> botocore.config.Config:
    return botocore.config.Config(
        retries={"max_attempts": 5},
        connect_timeout=10,
        max_pool_connections=10,
        user_agent_extra=f"awsorbit/{__version__}",
    )


def boto3_client(service_name: str) -> boto3.client:
    return boto3.Session().client(service_name=service_name, use_ssl=True, config=get_botocore_config())


def boto3_resource(service_name: str) -> boto3.client:
    return boto3.Session().resource(service_name=service_name, use_ssl=True, config=get_botocore_config())


def get_region() -> str:
    session = boto3.Session()
    if session.region_name is None:
        raise ValueError("It is not possible to infer AWS REGION from your environment.")
    return str(session.region_name)


def get_account_id() -> str:
    return str(boto3_client(service_name="sts").get_caller_identity().get("Account"))


def namedtuple_to_dict(obj: Any) -> Any:
    if hasattr(obj, "_asdict"):  # NamedTuple
        return dict(zip(obj._fields, (namedtuple_to_dict(item) for item in obj)))
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, dict):
        return dict(zip(obj.keys(), (namedtuple_to_dict(item) for item in obj.values())))
    elif isinstance(obj, Iterable):
        return list((namedtuple_to_dict(item) for item in obj))
    return obj


def path_from_filename(filename: str) -> str:
    return os.path.dirname(os.path.realpath(filename))


def upsert_subnet_tag(subnet_id: str, key: str, value: str) -> None:
    ec2: Any = boto3_resource("ec2")
    ec2.Subnet(subnet_id).create_tags(Tags=[{"Key": key, "Value": value}])


def replace_underscores(original: Dict[str, Any]) -> Dict[str, Any]:
    aux = {k.replace("_", "-"): v for k, v in original.items()}
    ret = {}
    for k, v in aux.items():
        if k.startswith("-"):
            ret[k[1:]] = v
        else:
            ret[k] = v
    return ret


def extract_plugin_module_name(func: Callable[..., Union[None, List[str], str]]) -> str:
    name = func.__module__.split(sep=".", maxsplit=1)[0]
    return name


def extract_images_names(env_name: str) -> List[str]:
    resp_type = Dict[str, List[Dict[str, List[Dict[str, str]]]]]
    try:
        response: resp_type = boto3_client("cloudformation").describe_stacks(StackName=f"orbit-{env_name}")
    except botocore.exceptions.ClientError as ex:
        error: Dict[str, Any] = ex.response["Error"]
        if error["Code"] == "ValidationError" and f"Stack with id orbit-{env_name} does not exist" in error["Message"]:
            return []
        raise
    if len(response["Stacks"]) < 1:
        return []
    if "Outputs" not in response["Stacks"][0]:
        return []
    for output in response["Stacks"][0]["Outputs"]:
        if output["ExportName"] == f"orbit-{env_name}-repos":
            _logger.debug("Export value: %s", output["OutputValue"])
            return output["OutputValue"].split(",")
    raise RuntimeError(f"Stack orbit-{env_name} does not have the expected orbit-{env_name}-repos output.")


def try_it(
    f: Callable[..., Any],
    ex: Any,
    base: float = 1.0,
    max_num_tries: int = 3,
    **kwargs: Any,
) -> Any:
    """Run function with decorrelated Jitter.

    Reference: https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """
    delay: float = base
    for i in range(max_num_tries):
        try:
            return f(**kwargs)
        except ex as exception:
            if i == (max_num_tries - 1):
                raise exception
            delay = random.uniform(base, delay * 3)
            _logger.error(
                "Retrying %s | Fail number %s/%s | Exception: %s",
                f,
                i + 1,
                max_num_tries,
                exception,
            )
            time.sleep(delay)


def get_dns_ip(context: "Context") -> str:
    """
    Reserved by AWS. The IP address of the DNS server is the base of the VPC network range plus two.
    For VPCs with multiple CIDR blocks, the IP address of the DNS server is located in the primary CIDR.
    We also reserve the base of each subnet range plus two for all CIDR blocks in the VPC.
    """
    if context.networking.vpc_cidr_block is None:
        raise ValueError("Impossible to localize the VPC CIDR Block!")
    base: str = context.networking.vpc_cidr_block[:-3]
    return str(ipaddress.ip_address(base) + 2)


def get_dns_ip_cidr(context: "Context") -> str:
    """
    Reserved by AWS. The IP address of the DNS server is the base of the VPC network range plus two.
    For VPCs with multiple CIDR blocks, the IP address of the DNS server is located in the primary CIDR.
    We also reserve the base of each subnet range plus two for all CIDR blocks in the VPC.
    """
    cidr: str = f"{get_dns_ip(context)}/32"
    _logger.debug("DNS CIDR: %s", cidr)
    return cidr


def print_dir(dir: str, exclude: List[str] = []) -> None:
    for root, dirnames, filenames in os.walk(dir):
        if exclude:
            for d in list(dirnames):
                if d in exclude:
                    dirnames.remove(d)
        # print path to all subdirectories first.
        for subdirname in dirnames:
            _logger.debug(os.path.join(root, subdirname))
        # print path to all filenames.
        for filename in filenames:
            _logger.debug((os.path.join(root, filename)))


def resolve_parameters(template: str, parameters: Dict[str, Optional[str]]) -> str:
    string_template = Template(template)
    template = string_template.safe_substitute(parameters)
    return template



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from typing import Any, Dict, List, Optional

import click

from aws_orbit import utils
from aws_orbit.messages import print_list, stylize
from aws_orbit.models.context import Context, ContextSerDe
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _fetch_repo_uri(names: List[str], context: "Context") -> Dict[str, str]:
    names = [f"orbit-{context.name}-{x}" for x in names]
    ret: Dict[str, str] = {x: "" for x in names}
    client = boto3_client("ecr")
    paginator = client.get_paginator("describe_repositories")
    for page in paginator.paginate(repositoryNames=names):
        for repo in page["repositories"]:
            ret[repo["repositoryName"]] = repo["repositoryUri"]
    ret = {k.replace(f"orbit-{context.name}-", ""): v for k, v in ret.items()}
    return ret


def list_images(env: str, region: Optional[str]) -> None:
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
    names = utils.extract_images_names(env_name=env)
    _logger.debug("names: %s", names)
    if names:
        uris = _fetch_repo_uri(names=names, context=context)
        print_list(
            tittle=f"Available docker images into the {stylize(context.name)} env:",
            items=[f"{k} {stylize(':')} {v}" for k, v in uris.items()],
        )
    else:
        click.echo(f"Thre is no docker images into the {stylize(context.name)} env.")


def list_env(env: str, variable: str) -> None:
    ssm = utils.boto3_client("ssm")
    res = ssm.get_parameters_by_path(Path="/orbit", Recursive=True)
    env_info: Dict[str, Any] = {}
    if env and len(env) > 0:
        _logger.debug(f"looking for {env}")
    while True:
        params = res["Parameters"]
        for p in params:
            if not p["Name"].endswith("context") or "teams" in p["Name"]:
                continue
            env_name = p["Name"].split("/")[2]
            if len(env) > 0 and not env_name == env:
                continue
            env_name = p["Name"].split("/")[2]
            context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
            _logger.debug(f"found env: {env_name}")
            if len(context.teams) > 0:
                teams_list: List[str] = [x.name for x in context.teams]
            else:
                teams_list = []

            if variable == "landing-page":
                print(context.landing_page_url)
            elif variable == "toolkitbucket":
                print(context.toolkit.s3_bucket)
            elif variable == "all":
                env_info[env_name] = {
                    "LandingPage": f"{context.landing_page_url}/orbit/login",
                    "Teams": teams_list,
                    "ToolkitBucket": context.toolkit.s3_bucket,
                }
            else:
                raise Exception(f"Unknown --variable option {variable}")

        if "NextToken" in res:
            res = ssm.get_parameters_by_path(Path="/orbit", Recursive=True, NextToken=res["NextToken"])
        else:
            break

    if variable == "all":
        if len(env_info) == 0:
            click.echo("There are no Orbit environments available")
            return
        else:
            print("Available Orbit environments:")
            print(json.dumps(env_info, indent=4, sort_keys=True))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import List, Optional, cast

import aws_cdk.aws_ec2 as ec2
import aws_cdk.aws_efs as efs
import aws_cdk.aws_kms as kms
import aws_cdk.core as core


class EfsBuilder:
    @staticmethod
    def build_file_system(
        scope: core.Construct,
        name: str,
        efs_life_cycle: str,
        vpc: ec2.IVpc,
        efs_security_group: ec2.ISecurityGroup,
        subnets: List[ec2.ISubnet],
        team_kms_key: kms.Key,
    ) -> efs.FileSystem:
        fs_name: str = f"orbit-{name}-fs"
        efs_fs: efs.FileSystem = efs.FileSystem(
            scope=scope,
            id=fs_name,
            file_system_name=fs_name,
            vpc=vpc,
            encrypted=True,
            lifecycle_policy=efs.LifecyclePolicy[efs_life_cycle] if efs_life_cycle else None,
            performance_mode=efs.PerformanceMode.GENERAL_PURPOSE,
            throughput_mode=efs.ThroughputMode.BURSTING,
            security_group=efs_security_group,
            vpc_subnets=ec2.SubnetSelection(subnets=subnets),
            kms_key=cast(Optional[kms.IKey], team_kms_key),
        )
        return efs_fs

    @staticmethod
    def build_file_system_access_point(
        scope: core.Construct, team_name: str, shared_fs: efs.FileSystem, path: str, ap_name: str
    ) -> efs.AccessPoint:
        return efs.AccessPoint(
            scope=scope,
            id=ap_name,
            file_system=cast(efs.IFileSystem, shared_fs),
            path=f"/{team_name}/{path}",
            posix_user=efs.PosixUser(gid="100", uid="1000"),
            create_acl=efs.Acl(owner_gid="100", owner_uid="1000", permissions="770"),
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import time
from typing import Any, Dict, List, Optional, Union, cast

import requests
from jose import jwk, jwt
from jose.utils import base64url_decode

logger = logging.getLogger()
logger.setLevel(logging.INFO)

COGNITO_USER_POOL_ID: str = os.environ["COGNITO_USER_POOL_ID"]
COGNITO_USER_POOL_CLIENT_ID: str = os.environ["COGNITO_USER_POOL_CLIENT_ID"]
REGION: str = os.environ["REGION"]

_cognito_keys: Optional[List[Dict[str, str]]] = None


def _get_keys() -> List[Dict[str, str]]:
    global _cognito_keys
    if _cognito_keys is None:
        print("Fetching keys...")
        url: str = f"https://cognito-idp.{REGION}.amazonaws.com/{COGNITO_USER_POOL_ID}/.well-known/jwks.json"
        _cognito_keys = cast(List[Dict[str, str]], requests.get(url).json()["keys"])
    return _cognito_keys


def get_claims(token: str) -> Dict[str, Union[str, int]]:
    # get the kid from the headers prior to verification
    headers: Dict[str, str] = cast(Dict[str, str], jwt.get_unverified_headers(token=token))
    kid: str = headers["kid"]
    # search for the kid in the downloaded public keys
    for key in _get_keys():
        if kid == key["kid"]:
            # construct the public key
            public_key = jwk.construct(key_data=key)
            break
    else:
        raise ValueError("Public key not found in JWK.")
    # get the last two sections of the token,
    # message and signature (encoded in base64)
    message, encoded_signature = token.rsplit(".", 1)
    # decode the signature
    decoded_signature: bytes = base64url_decode(encoded_signature.encode("utf-8"))
    # verify the signature
    if public_key.verify(msg=message.encode("utf8"), sig=decoded_signature) is False:
        raise RuntimeError("Signature verification failed.")
    print("Signature validaded.")
    # since we passed the verification, we can now safely use the unverified claims
    claims: Dict[str, Union[str, int]] = cast(Dict[str, Union[str, int]], jwt.get_unverified_claims(token))
    # additionally we can verify the token expiration
    if time.time() > int(claims["exp"]):
        raise ValueError("Token expired.")
    print("Token not expired.")
    # and the Audience (use claims['client_id'] if verifying an access token)
    if claims["aud"] != COGNITO_USER_POOL_CLIENT_ID:
        raise ValueError("Token was not issued for this audience.")
    # now we can use the claims
    return claims


def handler(event: Dict[str, Any], context: Optional[Dict[str, Any]]) -> Dict[str, Union[str, int]]:
    claims: Dict[str, Union[str, int]] = get_claims(token=event["token"])
    logger.info("claims: %s", claims)
    return claims



def get_users_url(user_pool_id: str, region: str) -> str:
    return f"https://{region}.console.aws.amazon.com/cognito/users/?region={region}#/pool/{user_pool_id}/users"


def get_pool_arn(user_pool_id: str, region: str, account: str) -> str:
    return f"arn:aws:cognito-idp:{region}:{account}:userpool/{user_pool_id}"



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import time
from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, NamedTuple, Optional, Union

import botocore.exceptions
import yaml

from aws_orbit import __version__
from aws_orbit.models.context import Context, FoundationContext
from aws_orbit.utils import boto3_client, try_it

if TYPE_CHECKING:
    from aws_orbit.models.changeset import Changeset

_logger: logging.Logger = logging.getLogger(__name__)


_BUILD_WAIT_POLLING_DELAY: float = 5  # SECONDS

CDK_VERSION = "~=1.100.0"
CDK_MODULES = [
    "aws_cdk.core",
    "aws-cdk.aws-ec2",
    "aws-cdk.aws-s3",
    "aws-cdk.aws-iam",
    "aws-cdk.aws-efs",
    "aws-cdk.aws-ecr",
    "aws-cdk.aws-ecs",
    "aws-cdk.aws-ssm",
    "aws-cdk.aws-kms",
    "aws-cdk.aws-cognito",
    "aws-cdk.aws-lambda",
    "aws-cdk.aws-lambda-python",
    "aws-cdk.aws-stepfunctions",
    "aws-cdk.aws-stepfunctions-tasks",
]


class BuildStatus(Enum):
    failed = "FAILED"
    fault = "FAULT"
    in_progress = "IN_PROGRESS"
    stopped = "STOPPED"
    succeeded = "SUCCEEDED"
    timed_out = "TIMED_OUT"


class BuildPhaseType(Enum):
    build = "BUILD"
    completed = "COMPLETED"
    download_source = "DOWNLOAD_SOURCE"
    finalizing = "FINALIZING"
    install = "INSTALL"
    post_build = "POST_BUILD"
    pre_build = "PRE_BUILD"
    provisioning = "PROVISIONING"
    queued = "QUEUED"
    submitted = "SUBMITTED"
    upload_artifacts = "UPLOAD_ARTIFACTS"


class BuildPhaseStatus(Enum):
    failed = "FAILED"
    fault = "FAULT"
    queued = "QUEUED"
    in_progress = "IN_PROGRESS"
    stopped = "STOPPED"
    succeeded = "SUCCEEDED"
    timed_out = "TIMED_OUT"


class BuildPhaseContext(NamedTuple):
    status_code: Optional[str]
    message: Optional[str]


class BuildPhase(NamedTuple):
    phase_type: BuildPhaseType
    status: Optional[BuildPhaseStatus]
    start_time: datetime
    end_time: Optional[datetime]
    duration_in_seconds: float
    contexts: List[BuildPhaseContext]


class BuildCloudWatchLogs(NamedTuple):
    enabled: bool
    group_name: Optional[str]
    stream_name: Optional[str]


class BuildInfo(NamedTuple):
    build_id: str
    status: BuildStatus
    current_phase: BuildPhaseType
    start_time: datetime
    end_time: Optional[datetime]
    duration_in_seconds: float
    phases: List[BuildPhase]
    logs: BuildCloudWatchLogs


def start(
    context: "Context",
    project_name: str,
    stream_name: str,
    bundle_location: str,
    buildspec: Dict[str, Any],
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
) -> str:
    client = boto3_client("codebuild")
    repo: Optional[str] = None
    credentials: Optional[str] = None
    if context.images.code_build.get_source(account_id=context.account_id, region=context.region).startswith("ecr"):
        version = context.images.code_build.version
        repo = f"{context.images.code_build.repository}:{version}"
        if not any(match in repo for match in [".amazonaws.com/", "public.ecr.aws"]):
            repo = f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com/{repo}:{version}"
        credentials = "SERVICE_ROLE"
    _logger.debug("Repository: %s", repo)
    _logger.debug("Credentials: %s", credentials)
    build_params = {
        "projectName": project_name,
        "sourceTypeOverride": "S3",
        "sourceLocationOverride": bundle_location,
        "buildspecOverride": yaml.safe_dump(data=buildspec, sort_keys=False, indent=4),
        "timeoutInMinutesOverride": timeout,
        "privilegedModeOverride": True,
        "logsConfigOverride": {
            "cloudWatchLogs": {
                "status": "ENABLED",
                "groupName": f"/aws/codebuild/{project_name}",
                "streamName": stream_name,
            },
            "s3Logs": {"status": "DISABLED"},
        },
    }
    if repo:
        build_params["imageOverride"] = repo
    if credentials:
        build_params["imagePullCredentialsTypeOverride"] = credentials

    if overrides:
        build_params = {**build_params, **overrides}

    response: Dict[str, Any] = client.start_build(**build_params)
    return str(response["build"]["id"])


def fetch_build_info(build_id: str) -> BuildInfo:
    client = boto3_client("codebuild")
    response: Dict[str, List[Dict[str, Any]]] = try_it(
        f=client.batch_get_builds, ex=botocore.exceptions.ClientError, ids=[build_id], max_num_tries=5
    )
    if not response["builds"]:
        raise RuntimeError(f"CodeBuild build {build_id} not found.")
    build = response["builds"][0]
    now = datetime.now(timezone.utc)
    log_enabled = True if build.get("logs", {}).get("cloudWatchLogs", {}).get("status") == "ENABLED" else False
    return BuildInfo(
        build_id=build_id,
        status=BuildStatus(value=build["buildStatus"]),
        current_phase=BuildPhaseType(value=build["currentPhase"]),
        start_time=build["startTime"],
        end_time=build.get("endTime", now),
        duration_in_seconds=(build.get("endTime", now) - build["startTime"]).total_seconds(),
        phases=[
            BuildPhase(
                phase_type=BuildPhaseType(value=p["phaseType"]),
                status=None if "phaseStatus" not in p else BuildPhaseStatus(value=p["phaseStatus"]),
                start_time=p["startTime"],
                end_time=p.get("endTime", now),
                duration_in_seconds=p.get("durationInSeconds"),
                contexts=[
                    BuildPhaseContext(status_code=p.get("statusCode"), message=p.get("message"))
                    for c in p.get("contexts", [])
                ],
            )
            for p in build["phases"]
        ],
        logs=BuildCloudWatchLogs(
            enabled=log_enabled,
            group_name=build["logs"]["cloudWatchLogs"].get("groupName") if log_enabled else None,
            stream_name=build["logs"]["cloudWatchLogs"].get("streamName") if log_enabled else None,
        ),
    )


def wait(build_id: str) -> Iterable[BuildInfo]:
    build = fetch_build_info(build_id=build_id)
    while build.status is BuildStatus.in_progress:
        time.sleep(_BUILD_WAIT_POLLING_DELAY)

        last_phase = build.current_phase
        last_status = build.status
        build = fetch_build_info(build_id=build_id)

        if build.current_phase is not last_phase or build.status is not last_status:
            _logger.debug("phase: %s (%s)", build.current_phase.value, build.status.value)

        yield build

    if build.status is not BuildStatus.succeeded:
        raise RuntimeError(f"CodeBuild build ({build_id}) is {build.status.value}")
    _logger.debug(
        "start: %s | end: %s | elapsed: %s",
        build.start_time,
        build.end_time,
        build.duration_in_seconds,
    )


SPEC_TYPE = Dict[str, Union[float, Dict[str, Dict[str, Union[List[str], Dict[str, float]]]]]]


def generate_spec(
    context: Union[Context, FoundationContext],
    changeset: Optional["Changeset"] = None,
    plugins: bool = True,
    cmds_install: Optional[List[str]] = None,
    cmds_pre: Optional[List[str]] = None,
    cmds_build: Optional[List[str]] = None,
    cmds_post: Optional[List[str]] = None,
) -> SPEC_TYPE:
    pre: List[str] = [] if cmds_pre is None else cmds_pre
    build: List[str] = [] if cmds_build is None else cmds_build
    post: List[str] = [] if cmds_post is None else cmds_post
    install = [
        (
            "nohup /usr/sbin/dockerd --host=unix:///var/run/docker.sock"
            " --host=tcp://0.0.0.0:2375 --storage-driver=overlay&"
        ),
        'timeout 15 sh -c "until docker info; do echo .; sleep 1; done"',
    ]

    if context.codeartifact_domain and context.codeartifact_repository:
        install.append(
            "aws codeartifact login --tool pip "
            f"--domain {context.codeartifact_domain} "
            f"--repository {context.codeartifact_repository}"
        )
        # Adding Codeartifact based pip.conf, later used for Dockerfile build.
        install.append("cp ~/.config/pip/pip.conf .")

    install.append("pwd")
    # Orbit Workbench CLI
    if __version__.endswith(".dev0"):
        install.append(f"pip install aws-orbit=={__version__}")
    else:
        install.append(f"pip install aws-orbit~={__version__}")

    # Login with any store Docker Registry credentials
    install.append(
        "/var/scripts/retrieve_docker_creds.py && echo 'Docker logins successful' || echo 'Docker logins failed'"
    )

    # Plugins
    if plugins and isinstance(context, Context):
        for team_context in context.teams:
            for plugin in team_context.plugins:
                if plugin.module is not None:
                    plugin_module_name = (plugin.module).replace("_", "-")
                    install.append(f"pip install --upgrade aws-orbit-{plugin_module_name}=={__version__}")

        if changeset is not None:
            for plugin_changeset in changeset.plugin_changesets:

                # OLD
                for plugin_name in plugin_changeset.old:
                    module: str = plugin_changeset.old_modules[plugin_name]
                    if plugin_name not in plugin_changeset.new and module is not None:
                        plugin_module_name = (module).replace("_", "-")
                        install.append(f"pip install --upgrade aws-orbit-{plugin_module_name}=={__version__}")

                # OLD
                for plugin_name in plugin_changeset.new:
                    module = plugin_changeset.new_modules[plugin_name]
                    if plugin_name not in plugin_changeset.old and module is not None:
                        plugin_module_name = (module).replace("_", "-")
                        install.append(f"pip install --upgrade aws-orbit-{plugin_module_name}=={__version__}")

    if cmds_install is not None:
        install += cmds_install

    return_spec: SPEC_TYPE = {
        "version": 0.2,
        "phases": {
            "install": {
                "runtime-versions": {"python": 3.7, "nodejs": 12, "docker": 19},
                "commands": install,
            },
            "pre_build": {"commands": pre},
            "build": {"commands": build},
            "post_build": {"commands": post},
        },
    }
    _logger.debug(return_spec)
    return return_spec



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import shutil
import sys
from typing import Any, Dict, List, cast

import jsii
from aws_cdk import aws_cognito as cognito
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_iam as iam
from aws_cdk import aws_kms as kms
from aws_cdk import aws_s3 as s3
from aws_cdk import aws_ssm as ssm
from aws_cdk import core
from aws_cdk.core import App, Aspects, CfnOutput, Construct, Duration, Stack, Tags

from aws_orbit.models.context import ContextSerDe, FoundationContext
from aws_orbit.remote_files.cdk.team_builders.codeartifact import DeployCodeArtifact
from aws_orbit.remote_files.cdk.team_builders.efs import EfsBuilder
from aws_orbit.remote_files.cdk.team_builders.s3 import S3Builder

_logger: logging.Logger = logging.getLogger(__name__)


class FoundationStack(Stack):
    def __init__(
        self, scope: Construct, id: str, context: "FoundationContext", ssl_cert_arn: str, **kwargs: Any
    ) -> None:
        self.env_name = context.name
        self.context = context
        self.ssl_cert_arn = ssl_cert_arn
        super().__init__(scope, id, **kwargs)
        Tags.of(scope=cast(core.IConstruct, self)).add(key="Env", value=f"orbit-{self.env_name}")
        self.vpc: ec2.Vpc = self._create_vpc(context)

        self.public_subnets = (
            self.vpc.select_subnets(subnet_type=ec2.SubnetType.PUBLIC)
            if self.vpc.public_subnets
            else self.vpc.select_subnets(subnet_name="")
        )
        self.private_subnets = (
            self.vpc.select_subnets(subnet_type=ec2.SubnetType.PRIVATE)
            if self.vpc.private_subnets
            else self.vpc.select_subnets(subnet_name="")
        )
        if not context.networking.data.internet_accessible:
            self.isolated_subnets = (
                self.vpc.select_subnets(subnet_type=ec2.SubnetType.ISOLATED)
                if self.vpc.isolated_subnets
                else self.vpc.select_subnets(subnet_name="")
            )
            self.nodes_subnets = self.isolated_subnets
        else:
            self.nodes_subnets = self.private_subnets

        self._vpc_security_group = ec2.SecurityGroup(
            self, "vpc-sg", vpc=cast(ec2.IVpc, self.vpc), allow_all_outbound=False
        )
        # Adding ingress rule to VPC CIDR
        self._vpc_security_group.add_ingress_rule(
            peer=ec2.Peer.ipv4(self.vpc.vpc_cidr_block), connection=ec2.Port.all_tcp()
        )

        if not context.networking.data.internet_accessible:
            self._create_vpc_endpoints()

        if context.toolkit.s3_bucket is None:
            raise ValueError("context.toolkit_s3_bucket is not defined")
        toolkit_s3_bucket_name: str = context.toolkit.s3_bucket
        acct: str = core.Aws.ACCOUNT_ID
        self.bucket_names: Dict[str, Any] = {
            "scratch-bucket": f"orbit-f-{self.env_name}-scratch-{acct}-{context.toolkit.deploy_id}",
            "toolkit-bucket": toolkit_s3_bucket_name,
        }
        self._build_kms_key_for_env()
        self.scratch_bucket: s3.Bucket = S3Builder.build_s3_bucket(
            scope=self,
            id="scratch_bucket",
            name=self.bucket_names["scratch-bucket"],
            scratch_retention_days=30,
            kms_key=self.env_kms_key,
        )

        self.efs_fs = EfsBuilder.build_file_system(
            scope=self,
            name="orbit-fs",
            efs_life_cycle="AFTER_7_DAYS",
            vpc=cast(ec2.IVpc, self.vpc),
            efs_security_group=cast(ec2.ISecurityGroup, self._vpc_security_group),
            subnets=self.nodes_subnets.subnets,
            team_kms_key=self.env_kms_key,
        )

        self.user_pool: cognito.UserPool = self._create_user_pool()

        # Checks if CodeArtifact exists outside of the scope of Orbit, else creates it.
        if self.context.codeartifact_domain and self.context.codeartifact_repository:
            self.domain_name = self.context.codeartifact_domain
            self.repository_name = self.context.codeartifact_repository
        else:
            self.codeartifact = DeployCodeArtifact(self, id="CodeArtifact-from-Fndn")
            self.domain_name = self.codeartifact.artifact_domain.domain_name
            self.repository_name = self.codeartifact.pypi_repo.repository_name

        self._ssm_parameter = ssm.StringParameter(
            self,
            id="/orbit/DemoParams",
            string_value=json.dumps(
                {
                    "VpcId": self.vpc.vpc_id,
                    "PublicSubnets": self.public_subnets.subnet_ids,
                    "PrivateSubnets": self.private_subnets.subnet_ids,
                    "IsolatedSubnets": self.isolated_subnets.subnet_ids
                    if not context.networking.data.internet_accessible
                    else [],
                    "NodesSubnets": self.nodes_subnets.subnet_ids,
                    "LoadBalancersSubnets": self.public_subnets.subnet_ids,
                    "KMSKey": self.env_kms_key.key_arn,
                    "SharedEfsFsId": self.efs_fs.file_system_id,
                    "ScratchBucketArn": self.scratch_bucket.bucket_arn,
                    "ScratchBucketName": self.scratch_bucket.bucket_name,
                    "UserPoolId": self.user_pool.user_pool_id,
                    "SharedEfsSgId": self._vpc_security_group.security_group_id,
                    "UserPoolProviderName": self.user_pool.user_pool_provider_name,
                    "SslCertArn": self.ssl_cert_arn,
                    "CodeartifactDomain": self.domain_name,
                    "CodeartifactRepository": self.repository_name,
                    "IsCodeartifactExternal": True
                    if self.context.codeartifact_domain and self.context.codeartifact_repository
                    else False,
                }
            ),
            type=ssm.ParameterType.STRING,
            description="Orbit Workbench Demo resources.",
            parameter_name=context.resources_ssm_parameter_name,
            simple_name=False,
            tier=ssm.ParameterTier.INTELLIGENT_TIERING,
        )

        CfnOutput(
            scope=self,
            id=f"{id}vpcid",
            export_name=f"orbit-f-{self.env_name}-vpc-id",
            value=self.vpc.vpc_id,
        )

        CfnOutput(
            scope=self,
            id=f"{id}publicsubnetsids",
            export_name=f"orbit-f-{self.env_name}-public-subnet-ids",
            value=",".join(self.public_subnets.subnet_ids),
        )
        CfnOutput(
            scope=self,
            id=f"{id}privatesubnetsids",
            export_name=f"orbit-f-{self.env_name}-private-subnet-ids",
            value=",".join(self.private_subnets.subnet_ids),
        )
        if not context.networking.data.internet_accessible:
            CfnOutput(
                scope=self,
                id=f"{id}isolatedsubnetsids",
                export_name=f"orbit-f-{self.env_name}-isolated-subnet-ids",
                value=",".join(self.isolated_subnets.subnet_ids),
            )
        CfnOutput(
            scope=self,
            id=f"{id}nodesubnetsids",
            export_name=f"orbit-f-{self.env_name}-nodes-subnet-ids",
            value=",".join(self.nodes_subnets.subnet_ids),
        )

    def _build_kms_key_for_env(self) -> None:
        administrator_arns: List[str] = []  # A place to add other admins if needed for KMS
        admin_principals = iam.CompositePrincipal(
            *[iam.ArnPrincipal(arn) for arn in administrator_arns],
            iam.ArnPrincipal(f"arn:aws:iam::{self.context.account_id}:root"),
        )
        self.env_kms_key: kms.Key = kms.Key(
            self,
            id="kms-key",
            removal_policy=core.RemovalPolicy.RETAIN,
            enabled=True,
            enable_key_rotation=True,
            policy=iam.PolicyDocument(
                statements=[
                    iam.PolicyStatement(
                        effect=iam.Effect.ALLOW,
                        actions=["kms:*"],
                        resources=["*"],
                        principals=[cast(iam.IPrincipal, admin_principals)],
                    )
                ]
            ),
        )

    def _create_vpc(self, context: "FoundationContext") -> ec2.Vpc:
        if context.networking.data.internet_accessible:
            subnet_configuration = [
                ec2.SubnetConfiguration(name="Public", subnet_type=ec2.SubnetType.PUBLIC, cidr_mask=24),
                ec2.SubnetConfiguration(name="Private", subnet_type=ec2.SubnetType.PRIVATE, cidr_mask=21),
            ]
        else:
            subnet_configuration = [
                ec2.SubnetConfiguration(name="Public", subnet_type=ec2.SubnetType.PUBLIC, cidr_mask=24),
                ec2.SubnetConfiguration(name="Private", subnet_type=ec2.SubnetType.PRIVATE, cidr_mask=21),
                ec2.SubnetConfiguration(name="Isolated", subnet_type=ec2.SubnetType.ISOLATED, cidr_mask=21),
            ]

        vpc = ec2.Vpc(
            scope=self,
            id="vpc",
            default_instance_tenancy=ec2.DefaultInstanceTenancy.DEFAULT,
            cidr="10.0.0.0/16",
            enable_dns_hostnames=True,
            enable_dns_support=True,
            max_azs=context.networking.max_availability_zones
            if context.networking.max_availability_zones is not None
            else 2,
            nat_gateways=1,
            subnet_configuration=subnet_configuration,
        )
        return vpc

    def _create_user_pool(self) -> cognito.UserPool:
        pool = cognito.UserPool(
            scope=self,
            id="orbit-user-pool",
            account_recovery=cognito.AccountRecovery.EMAIL_ONLY,
            auto_verify=cognito.AutoVerifiedAttrs(email=True, phone=False),
            custom_attributes=None,
            email_settings=None,
            lambda_triggers=None,
            mfa=cognito.Mfa.OFF,
            mfa_second_factor=None,
            password_policy=cognito.PasswordPolicy(
                min_length=8,
                require_digits=True,
                require_lowercase=True,
                require_symbols=True,
                require_uppercase=True,
                temp_password_validity=Duration.days(5),
            ),
            self_sign_up_enabled=False,
            sign_in_aliases=cognito.SignInAliases(email=True, phone=False, preferred_username=False, username=True),
            sign_in_case_sensitive=True,
            sms_role=None,
            sms_role_external_id=None,
            standard_attributes=cognito.StandardAttributes(
                email=cognito.StandardAttribute(required=True, mutable=True)
            ),
            user_invitation=cognito.UserInvitationConfig(
                email_subject="Invite to join Orbit Workbench!",
                email_body="Hello, you have been invited to join Orbit Workbench!<br/><br/>"
                "Username: {username}<br/>"
                "Temporary password: {####}<br/><br/>"
                "Regards",
            ),
            user_pool_name=f"orbit-{self.env_name}-user-pool",
        )
        pool.apply_removal_policy(policy=core.RemovalPolicy.DESTROY)
        pool.add_domain(
            id="orbit-user-pool-domain",
            cognito_domain=cognito.CognitoDomainOptions(
                domain_prefix=f"orbit-{self.context.account_id}-{self.env_name}"
            ),
        )
        return pool

    def _create_vpc_endpoints(self) -> None:
        vpc_gateway_endpoints = {
            "s3": ec2.GatewayVpcEndpointAwsService.S3,
            "dynamodb": ec2.GatewayVpcEndpointAwsService.DYNAMODB,
        }
        vpc_interface_endpoints = {
            "cloudwatch_endpoint": ec2.InterfaceVpcEndpointAwsService.CLOUDWATCH,
            "cloudwatch_logs_endpoint": ec2.InterfaceVpcEndpointAwsService.CLOUDWATCH_LOGS,
            "cloudwatch_events": ec2.InterfaceVpcEndpointAwsService.CLOUDWATCH_EVENTS,
            "ecr_docker_endpoint": ec2.InterfaceVpcEndpointAwsService.ECR_DOCKER,
            "ecr_endpoint": ec2.InterfaceVpcEndpointAwsService.ECR,
            "ec2_endpoint": ec2.InterfaceVpcEndpointAwsService.EC2,
            "ecs": ec2.InterfaceVpcEndpointAwsService.ECS,
            "ecs_agent": ec2.InterfaceVpcEndpointAwsService.ECS_AGENT,
            "ecs_telemetry": ec2.InterfaceVpcEndpointAwsService.ECS_TELEMETRY,
            "git_endpoint": ec2.InterfaceVpcEndpointAwsService.CODECOMMIT_GIT,
            "ssm_endpoint": ec2.InterfaceVpcEndpointAwsService.SSM,
            "ssm_messages_endpoint": ec2.InterfaceVpcEndpointAwsService.SSM_MESSAGES,
            "secrets_endpoint": ec2.InterfaceVpcEndpointAwsService.SECRETS_MANAGER,
            "kms_endpoint": ec2.InterfaceVpcEndpointAwsService.KMS,
            "sagemaker_endpoint": ec2.InterfaceVpcEndpointAwsService.SAGEMAKER_API,
            "sagemaker_runtime": ec2.InterfaceVpcEndpointAwsService.SAGEMAKER_RUNTIME,
            "notebook_endpoint": ec2.InterfaceVpcEndpointAwsService.SAGEMAKER_NOTEBOOK,
            "athena_endpoint": ec2.InterfaceVpcEndpointAwsService("athena"),
            "glue_endpoint": ec2.InterfaceVpcEndpointAwsService("glue"),
            "sqs": ec2.InterfaceVpcEndpointAwsService.SQS,
            "step_function_endpoint": ec2.InterfaceVpcEndpointAwsService("states"),
            "sns_endpoint": ec2.InterfaceVpcEndpointAwsService.SNS,
            "kinesis_firehose_endpoint": ec2.InterfaceVpcEndpointAwsService("kinesis-firehose"),
            "api_gateway": ec2.InterfaceVpcEndpointAwsService.APIGATEWAY,
            "sts_endpoint": ec2.InterfaceVpcEndpointAwsService.STS,
            "efs": ec2.InterfaceVpcEndpointAwsService.ELASTIC_FILESYSTEM,
            "elb": ec2.InterfaceVpcEndpointAwsService.ELASTIC_LOAD_BALANCING,
            "autoscaling": ec2.InterfaceVpcEndpointAwsService("autoscaling"),
            "cloudformation_endpoint": ec2.InterfaceVpcEndpointAwsService("cloudformation"),
            "codebuild_endpoint": ec2.InterfaceVpcEndpointAwsService("codebuild"),
            "emr-containers": ec2.InterfaceVpcEndpointAwsService("emr-containers"),
            "databrew": ec2.InterfaceVpcEndpointAwsService("databrew"),
        }

        for name, gateway_vpc_endpoint_service in vpc_gateway_endpoints.items():
            self.vpc.add_gateway_endpoint(
                id=name,
                service=gateway_vpc_endpoint_service,
                subnets=[
                    ec2.SubnetSelection(subnets=self.nodes_subnets.subnets),
                ],
            )

        for name, interface_service in vpc_interface_endpoints.items():
            self.vpc.add_interface_endpoint(
                id=name,
                service=interface_service,
                subnets=ec2.SubnetSelection(subnets=self.nodes_subnets.subnets),
                private_dns_enabled=True,
                security_groups=[cast(ec2.ISecurityGroup, self._vpc_security_group)],
            )
        # Adding CodeArtifact VPC endpoints
        self.vpc.add_interface_endpoint(
            id="code_artifact_repo_endpoint",
            service=cast(
                ec2.IInterfaceVpcEndpointService, ec2.InterfaceVpcEndpointAwsService("codeartifact.repositories")
            ),
            subnets=ec2.SubnetSelection(subnets=self.nodes_subnets.subnets),
            private_dns_enabled=False,
            security_groups=[cast(ec2.ISecurityGroup, self._vpc_security_group)],
        )
        self.vpc.add_interface_endpoint(
            id="code_artifact_api_endpoint",
            service=cast(ec2.IInterfaceVpcEndpointService, ec2.InterfaceVpcEndpointAwsService("codeartifact.api")),
            subnets=ec2.SubnetSelection(subnets=self.nodes_subnets.subnets),
            private_dns_enabled=False,
            security_groups=[cast(ec2.ISecurityGroup, self._vpc_security_group)],
        )

        # Adding Lambda and Redshift endpoints with CDK low level APIs
        endpoint_url_template = "com.amazonaws.{}.{}"
        ec2.CfnVPCEndpoint(
            self,
            "redshift_endpoint",
            vpc_endpoint_type="Interface",
            service_name=endpoint_url_template.format(self.region, "redshift"),
            vpc_id=self.vpc.vpc_id,
            security_group_ids=[self._vpc_security_group.security_group_id],
            subnet_ids=self.nodes_subnets.subnet_ids,
            private_dns_enabled=True,
        )
        ec2.CfnVPCEndpoint(
            self,
            "lambda_endpoint",
            vpc_endpoint_type="Interface",
            service_name=endpoint_url_template.format(self.region, "lambda"),
            vpc_id=self.vpc.vpc_id,
            security_group_ids=[self._vpc_security_group.security_group_id],
            subnet_ids=self.nodes_subnets.subnet_ids,
            private_dns_enabled=True,
        )


def main() -> None:
    _logger.debug("sys.argv: %s", sys.argv)
    context: "FoundationContext" = ContextSerDe.load_context_from_ssm(env_name=sys.argv[1], type=FoundationContext)
    ssl_cert_arn: str
    if len(sys.argv) == 3:
        ssl_cert_arn = sys.argv[2]
    elif len(sys.argv) == 2:
        ssl_cert_arn = ""
    else:
        raise ValueError("Unexpected number of values in sys.argv.")

    outdir = os.path.join(".orbit.out", context.name, "cdk", cast(str, context.stack_name))
    os.makedirs(outdir, exist_ok=True)
    shutil.rmtree(outdir)

    app = App(
        outdir=outdir,
    )

    @jsii.implements(core.IAspect)
    class AddDeployPathIAM:
        """ Implementing CDK Aspects to add optional IAM Role prefix to IAM roles """

        def visit(self, node: core.IConstruct) -> None:
            """ Function to implement a path pattern """
            if isinstance(node, iam.CfnRole):
                node.path = f"/{context.role_prefix}/" if context.role_prefix else "/"

    foundation_stack = FoundationStack(
        scope=app,
        id=cast(str, context.stack_name),
        context=context,
        ssl_cert_arn=ssl_cert_arn,
        env=core.Environment(account=os.environ["CDK_DEFAULT_ACCOUNT"], region=os.environ["CDK_DEFAULT_REGION"]),
    )

    Aspects.of(scope=cast(core.IConstruct, foundation_stack)).add(cast(core.IAspect, AddDeployPathIAM()))
    app.synth(force=True)


if __name__ == "__main__":
    main()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# flake8: noqa: F811

import concurrent.futures
import json
import logging
from enum import IntEnum, auto
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Generic, List, Optional, Type, TypeVar, Union, cast

import botocore.exceptions
from dataclasses import field
from marshmallow import Schema
from marshmallow_dataclass import dataclass

import aws_orbit
from aws_orbit import utils
from aws_orbit.models.common import BaseSchema
from aws_orbit.models.manifest import (
    DataNetworkingManifest,
    FoundationImagesManifest,
    FoundationManifest,
    FrontendNetworkingManifest,
    ImagesManifest,
    ManagedNodeGroupManifest,
    Manifest,
    NetworkingManifest,
    PluginManifest,
    TeamManifest,
)
from aws_orbit.services import ssm
from aws_orbit.utils import boto3_client, boto3_resource

_logger: logging.Logger = logging.getLogger(__name__)


def get_container_defaults() -> Dict[str, int]:
    return {"cpu": 4, "memory": 16384}


class SubnetKind(IntEnum):
    private: int = auto()
    public: int = auto()
    isolated: int = auto()


@dataclass(base_schema=BaseSchema)
class SubnetContext:
    Schema: ClassVar[Type[Schema]] = Schema
    subnet_id: str
    kind: SubnetKind
    vpc_id: str
    cidr_block: Optional[str] = None
    availability_zone: Optional[str] = None
    route_table_id: Optional[str] = None

    def _fetch_route_table_id(self) -> None:
        ec2_client = boto3_client("ec2")
        res: Dict[str, Any] = ec2_client.describe_route_tables(
            Filters=[{"Name": "association.subnet-id", "Values": [self.subnet_id]}]
        )
        for route_table in res["RouteTables"]:
            if "Associations" in route_table:
                for association in route_table["Associations"]:
                    if association["SubnetId"] == self.subnet_id:
                        self.route_table_id = association["RouteTableId"]
                        return

    def fetch_properties(self) -> None:
        try:
            ec2 = boto3_resource("ec2")
            subnet = ec2.Subnet(self.subnet_id)
            self.cidr_block = str(subnet.cidr_block)
            self.availability_zone = str(subnet.availability_zone)
            self.vpc_id = str(subnet.vpc_id)
            self._fetch_route_table_id()
            _logger.debug("Properties from subnet %s successfully fetched.", self.subnet_id)
        except botocore.exceptions.ClientError:
            _logger.debug("Unable to fetch properties from subnet (%s) right now.", self.subnet_id)


@dataclass(base_schema=BaseSchema)
class NetworkingContext:
    Schema: ClassVar[Type[Schema]] = Schema
    vpc_id: Optional[str] = None
    vpc_cidr_block: Optional[str] = None
    public_subnets: List[SubnetContext] = field(default_factory=list)
    private_subnets: List[SubnetContext] = field(default_factory=list)
    isolated_subnets: List[SubnetContext] = field(default_factory=list)
    availability_zones: Optional[List[str]] = None
    frontend: FrontendNetworkingManifest = FrontendNetworkingManifest()
    data: DataNetworkingManifest = DataNetworkingManifest()
    max_availability_zones: Optional[int] = None

    def _fetch_vpc_cidr(self) -> None:
        ec2 = boto3_resource("ec2")
        vpc = ec2.Vpc(self.vpc_id)
        self.vpc_cidr_block = str(vpc.cidr_block)

    def _fetch_subnets_properties(self) -> None:
        subnets = self.public_subnets + self.private_subnets + self.isolated_subnets
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(subnets)) as executor:
            list(executor.map(lambda s: cast(None, s.fetch_properties()), subnets))

    def fetch_properties(self) -> None:
        if self.vpc_id:
            self._fetch_vpc_cidr()
            self._fetch_subnets_properties()
            azs = sorted(
                list(
                    set(
                        s.availability_zone
                        for s in self.public_subnets + self.private_subnets + self.isolated_subnets
                        if s.availability_zone is not None
                    )
                )
            )
            self.availability_zones = azs if azs else None


@dataclass(base_schema=BaseSchema)
class TeamContext:
    Schema: ClassVar[Type[Schema]] = Schema

    # Manifest
    name: str
    policies: List[str]
    grant_sudo: bool
    jupyterhub_inbound_ranges: List[str]
    image: Optional[str]
    plugins: List[PluginManifest]
    efs_life_cycle: Optional[str]

    # Context
    base_image_address: str
    final_image_address: str
    stack_name: str
    ssm_parameter_name: str
    team_ssm_parameter_name: str
    bootstrap_s3_prefix: str
    scratch_bucket: Optional[str] = None
    container_defaults: Dict[str, Any] = field(default_factory=get_container_defaults)
    efs_id: Optional[str] = None
    efs_ap_id: Optional[str] = None
    efs_private_ap_id: Optional[str] = None
    eks_pod_role_arn: Optional[str] = None
    jupyter_url: Optional[str] = None
    ecs_cluster_name: Optional[str] = None
    container_runner_arn: Optional[str] = None
    eks_k8s_api_arn: Optional[str] = None
    team_kms_key_arn: Optional[str] = None
    elbs: Optional[Dict[str, Dict[str, Any]]] = None
    team_security_group_id: Optional[str] = None
    # added with defaults for backward compatability
    fargate: bool = True
    k8_admin: bool = False
    team_helm_repository: Optional[str] = None
    user_helm_repository: Optional[str] = None

    # Manifest default
    authentication_groups: Optional[List[str]] = None
    profiles: Optional[List[Dict[str, Union[str, Dict[str, Any]]]]] = None

    def fetch_team_data(self) -> None:
        _logger.debug("Fetching Team %s data...", self.name)
        values = ssm.get_parameter(name=self.team_ssm_parameter_name)
        self.efs_id = values["EfsId"]
        self.efs_ap_id = values["EfsApId"]
        self.efs_private_ap_id = values["EfsApIdPrivate"]
        self.eks_pod_role_arn = values["EksPodRoleArn"]
        self.scratch_bucket = values["ScratchBucket"]
        self.team_kms_key_arn = values["TeamKmsKeyArn"]
        self.team_security_group_id = values["TeamSecurityGroupId"]


@dataclass(base_schema=BaseSchema)
class ToolkitManifest:
    stack_name: str
    codebuild_project: str
    deploy_id: Optional[str] = None
    admin_role: Optional[str] = None
    admin_role_arn: Optional[str] = None
    kms_arn: Optional[str] = None
    kms_alias: Optional[str] = None
    s3_bucket: Optional[str] = None


@dataclass(base_schema=BaseSchema)
class CdkToolkitManifest:
    stack_name: str
    s3_bucket: Optional[str] = None


@dataclass(base_schema=BaseSchema)
class FoundationContext:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    account_id: str
    region: str
    env_tag: str
    toolkit: ToolkitManifest
    cdk_toolkit: CdkToolkitManifest
    networking: NetworkingContext = NetworkingContext()
    is_codeartifact_external: Optional[bool] = None
    user_pool_id: Optional[str] = None
    scratch_bucket_arn: Optional[str] = None
    cognito_users_url: Optional[str] = None
    cognito_external_provider_redirect: Optional[str] = None
    cognito_external_provider_domain: Optional[str] = None
    role_prefix: Optional[str] = None
    stack_name: Optional[str] = None
    ssm_parameter_name: Optional[str] = None
    resources_ssm_parameter_name: Optional[str] = None
    codeartifact_domain: Optional[str] = None
    codeartifact_repository: Optional[str] = None
    images: FoundationImagesManifest = FoundationImagesManifest()
    policies: Optional[List[str]] = cast(List[str], field(default_factory=list))
    orbit_version: Optional[str] = aws_orbit.__version__


@dataclass(base_schema=BaseSchema)
class Context:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    account_id: str
    region: str
    env_tag: str
    env_stack_name: str
    env_ssm_parameter_name: str
    eks_stack_name: str
    ssm_parameter_name: str
    ssm_dockerhub_parameter_name: str
    toolkit: ToolkitManifest
    cdk_toolkit: CdkToolkitManifest
    networking: NetworkingContext
    role_prefix: Optional[str] = None
    user_pool_id: Optional[str] = None
    scratch_bucket_arn: Optional[str] = None
    cognito_users_url: Optional[str] = None
    cognito_external_provider_redirect: Optional[str] = None
    cognito_external_provider_domain: Optional[str] = None
    landing_page_url: Optional[str] = None
    k8_dashboard_url: Optional[str] = None
    codeartifact_domain: Optional[str] = None
    codeartifact_repository: Optional[str] = None
    cognito_external_provider: Optional[str] = None
    cognito_external_provider_label: Optional[str] = None
    eks_cluster_role_arn: Optional[str] = None
    eks_fargate_profile_role_arn: Optional[str] = None
    eks_env_nodegroup_role_arn: Optional[str] = None
    eks_oidc_provider: Optional[str] = None
    eks_system_masters_roles: Optional[List[str]] = cast(List[str], field(default_factory=list))
    user_pool_client_id: Optional[str] = None
    identity_pool_id: Optional[str] = None
    teams: List[TeamContext] = field(default_factory=list)
    images: ImagesManifest = ImagesManifest()
    elbs: Optional[Dict[str, Dict[str, Any]]] = None
    shared_efs_fs_id: Optional[str] = None
    shared_efs_sg_id: Optional[str] = None
    cluster_sg_id: Optional[str] = None
    cluster_pod_sg_id: Optional[str] = None
    managed_nodegroups: List[ManagedNodeGroupManifest] = field(default_factory=list)
    policies: Optional[List[str]] = cast(List[str], field(default_factory=list))
    orbit_version: Optional[str] = aws_orbit.__version__
    helm_repository: Optional[str] = None
    install_ssm_agent: Optional[bool] = False
    install_image_replicator: Optional[bool] = False

    def get_team_by_name(self, name: str) -> Optional[TeamContext]:
        for t in self.teams:
            if t.name == name:
                return t
        return None

    def remove_team_by_name(self, name: str) -> None:
        self.teams = [t for t in self.teams if t.name != name]

    def fetch_env_data(self) -> None:
        _logger.debug("Fetching Env data...")
        values = ssm.get_parameter(name=self.env_ssm_parameter_name)
        self.eks_cluster_role_arn = values["EksClusterRoleArn"]
        self.eks_fargate_profile_role_arn = values["EksFargateProfileRoleArn"]
        self.eks_env_nodegroup_role_arn = values["EksEnvNodegroupRoleArn"]
        self.user_pool_id = values["UserPoolId"]
        self.user_pool_client_id = values["UserPoolClientId"]
        self.identity_pool_id = values["IdentityPoolId"]
        self.cluster_pod_sg_id = values["ClusterPodSecurityGroupId"]
        self.fetch_cognito_external_idp_data()
        _logger.debug("Env data fetched successfully.")

    def fetch_cognito_external_idp_data(self) -> None:
        _logger.debug("Fetching Cognito External IdP data...")
        client = boto3_client(service_name="cognito-idp")
        response: Dict[str, Any] = client.describe_user_pool(UserPoolId=self.user_pool_id)
        domain: str = response["UserPool"].get("Domain")
        self.cognito_external_provider_domain = f"{domain}.auth.{self.region}.amazoncognito.com"
        _logger.debug("cognito_external_provider_domain: %s", self.cognito_external_provider_domain)
        response = client.describe_user_pool_client(UserPoolId=self.user_pool_id, ClientId=self.user_pool_client_id)
        self.cognito_external_provider_redirect = response["UserPoolClient"]["CallbackURLs"][0]
        _logger.debug("cognito_external_provider_redirect: %s", self.cognito_external_provider_redirect)
        _logger.debug("Cognito External IdP data fetched successfully.")

    def fetch_teams_data(self) -> None:
        _logger.debug("Fetching Teams data...")
        for team in self.teams:
            team.fetch_team_data()
        self.fetch_cognito_external_idp_data()
        _logger.debug("Env data fetched successfully.")


def create_team_context_from_manifest(manifest: "Manifest", team_manifest: "TeamManifest") -> TeamContext:
    account_id: str = utils.get_account_id()
    region: str = utils.get_region()
    ssm_parameter_name: str = f"/orbit/{manifest.name}/teams/{team_manifest.name}/context"
    version = manifest.images.jupyter_user.version
    _logger.debug("Team Base Image: %s - %s", team_manifest.name, manifest.images.jupyter_user)
    base_image_address = (
        f"{account_id}.dkr.ecr.{region}.amazonaws.com/orbit-{manifest.name}/jupyter-user:{version}"
        if manifest.images.jupyter_user.get_source(account_id=account_id, region=region) == "code"
        else f"{manifest.images.jupyter_user.repository}:{version}"
    )
    final_image_address = base_image_address
    return TeamContext(
        base_image_address=base_image_address,
        final_image_address=final_image_address,
        stack_name=f"orbit-{manifest.name}-{team_manifest.name}",
        ssm_parameter_name=ssm_parameter_name,
        team_ssm_parameter_name=f"/orbit/{manifest.name}/teams/{team_manifest.name}/team",
        bootstrap_s3_prefix=f"teams/{manifest.name}/bootstrap/",
        **team_manifest.__dict__,
    )


def create_teams_context_from_manifest(manifest: "Manifest") -> List[TeamContext]:
    teams: List[TeamContext] = []
    for team in manifest.teams:
        ssm_parameter_name: str = f"/orbit/{manifest.name}/teams/{team.name}/context"
        if ssm.does_parameter_exist(name=ssm_parameter_name) is False:
            _logger.debug("Team %s is not deployed yet.", team.name)
            continue
        teams.append(create_team_context_from_manifest(manifest=manifest, team_manifest=team))
    return teams


def create_networking_context_from_manifest(networking: "NetworkingManifest") -> NetworkingContext:
    args: Dict[str, Any] = {
        "frontend": networking.frontend,
        "data": networking.data,
        "max_availability_zones": networking.max_availability_zones,
    }
    if networking.vpc_id:
        args["vpc_id"] = networking.vpc_id
        args["public_subnets"] = [
            SubnetContext(subnet_id=x, kind=SubnetKind.public, vpc_id=networking.vpc_id)
            for x in networking.public_subnets
        ]
        args["private_subnets"] = [
            SubnetContext(subnet_id=x, kind=SubnetKind.private, vpc_id=networking.vpc_id)
            for x in networking.private_subnets
        ]
        args["isolated_subnets"] = [
            SubnetContext(subnet_id=x, kind=SubnetKind.isolated, vpc_id=networking.vpc_id)
            for x in networking.isolated_subnets
        ]
    ctx = NetworkingContext(**args)
    ctx.fetch_properties()
    return ctx


T = TypeVar("T")
V = TypeVar("V")


class ContextSerDe(Generic[T, V]):
    @staticmethod
    def load_context_from_manifest(manifest: T) -> V:
        if isinstance(manifest, Manifest):
            return cast(V, ContextSerDe._load_context_from_manifest(manifest))
        elif isinstance(manifest, FoundationManifest):
            return cast(V, ContextSerDe._load_foundation_context_from_manifest(manifest))
        else:
            raise ValueError("Unknown 'manifest' Type")

    @staticmethod
    def _load_context_from_manifest(manifest: "Manifest") -> Context:
        _logger.debug("Loading Context from manifest")
        context_parameter_name: str = f"/orbit/{manifest.name}/context"
        if ssm.does_parameter_exist(name=context_parameter_name):
            context: Context = ContextSerDe.load_context_from_ssm(env_name=manifest.name, type=Context)
            context.networking = create_networking_context_from_manifest(networking=manifest.networking)
            context.user_pool_id = manifest.user_pool_id
            context.shared_efs_fs_id = manifest.shared_efs_fs_id
            context.shared_efs_sg_id = manifest.shared_efs_sg_id
            context.scratch_bucket_arn = manifest.scratch_bucket_arn
            context.policies = manifest.policies
            context.codeartifact_domain = manifest.codeartifact_domain
            context.codeartifact_repository = manifest.codeartifact_repository
            context.role_prefix = manifest.role_prefix
            context.cognito_external_provider = manifest.cognito_external_provider
            context.cognito_external_provider_label = manifest.cognito_external_provider_label

        else:
            context = Context(
                name=manifest.name,
                account_id=utils.get_account_id(),
                region=utils.get_region(),
                env_tag=f"orbit-{manifest.name}",
                env_stack_name=f"orbit-{manifest.name}",
                env_ssm_parameter_name=f"/orbit/{manifest.name}/env",
                eks_stack_name=f"eksctl-orbit-{manifest.name}-cluster",
                ssm_parameter_name=context_parameter_name,
                ssm_dockerhub_parameter_name=f"/orbit/{manifest.name}/dockerhub",
                toolkit=ToolkitManifest(
                    stack_name=f"orbit-{manifest.name}-toolkit", codebuild_project=f"orbit-{manifest.name}"
                ),
                cdk_toolkit=CdkToolkitManifest(stack_name=f"orbit-{manifest.name}-cdk-toolkit"),
                codeartifact_domain=manifest.codeartifact_domain,
                codeartifact_repository=manifest.codeartifact_repository,
                role_prefix=manifest.role_prefix,
                scratch_bucket_arn=manifest.scratch_bucket_arn,
                networking=create_networking_context_from_manifest(networking=manifest.networking),
                images=manifest.images,
                user_pool_id=manifest.user_pool_id,
                cognito_external_provider=manifest.cognito_external_provider,
                cognito_external_provider_label=manifest.cognito_external_provider_label,
                teams=create_teams_context_from_manifest(manifest=manifest),
                shared_efs_fs_id=manifest.shared_efs_fs_id,
                shared_efs_sg_id=manifest.shared_efs_sg_id,
                managed_nodegroups=[],
                eks_system_masters_roles=[],
                policies=manifest.policies,
                install_ssm_agent=manifest.install_ssm_agent,
                install_image_replicator=manifest.install_image_replicator,
            )
        context.install_ssm_agent = manifest.install_ssm_agent
        context.install_image_replicator = manifest.install_image_replicator

        ContextSerDe.fetch_toolkit_data(context=context)
        ContextSerDe.dump_context_to_ssm(context=context)
        return context

    @staticmethod
    def _load_foundation_context_from_manifest(manifest: "FoundationManifest") -> FoundationContext:
        _logger.debug("Loading Context from manifest")
        context_parameter_name: str = f"/orbit-f/{manifest.name}/context"
        if ssm.does_parameter_exist(name=context_parameter_name):
            context: FoundationContext = ContextSerDe.load_context_from_ssm(
                env_name=manifest.name, type=FoundationContext
            )
            context.images = manifest.images
            context.policies = manifest.policies
            context.codeartifact_domain = manifest.codeartifact_domain
            context.codeartifact_repository = manifest.codeartifact_repository
            context.role_prefix = manifest.role_prefix
            context.networking = create_networking_context_from_manifest(networking=manifest.networking)
        else:
            context = FoundationContext(
                name=manifest.name,
                account_id=utils.get_account_id(),
                region=utils.get_region(),
                env_tag=f"orbit-f-{manifest.name}",
                ssm_parameter_name=context_parameter_name,
                resources_ssm_parameter_name=f"/orbit-f/{manifest.name}/resources",
                toolkit=ToolkitManifest(
                    stack_name=f"orbit-f-{manifest.name}-toolkit",
                    codebuild_project=f"orbit-f-{manifest.name}",
                ),
                cdk_toolkit=CdkToolkitManifest(stack_name=f"orbit-f-{manifest.name}-cdk-toolkit"),
                codeartifact_domain=manifest.codeartifact_domain,
                codeartifact_repository=manifest.codeartifact_repository,
                role_prefix=manifest.role_prefix,
                networking=create_networking_context_from_manifest(networking=manifest.networking),
                images=manifest.images,
                policies=manifest.policies,
                stack_name=f"orbit-f-{manifest.name}",
            )
        ContextSerDe.fetch_toolkit_data(context=context)
        ContextSerDe.dump_context_to_ssm(context=context)
        return context

    @staticmethod
    def dump_context_to_str(context: V) -> str:
        if isinstance(context, Context):
            content: Dict[str, Any] = cast(Dict[str, Any], Context.Schema().dump(context))
        elif isinstance(context, FoundationContext):
            content = cast(Dict[str, Any], FoundationContext.Schema().dump(context))
        elif isinstance(context, TeamContext):
            content = cast(Dict[str, Any], TeamContext.Schema().dump(context))
        else:
            raise ValueError("Unknown 'context' Type")
        return str(json.dumps(obj=content, sort_keys=True))

    def dump_context_to_ssm(context: V) -> None:
        _logger.debug("Writing context to SSM parameter.")

        if isinstance(context, Context):
            _logger.debug("Teams: %s", [t.name for t in context.teams])
            content: Dict[str, Any] = cast(Dict[str, Any], Context.Schema().dump(context))
            current_teams_contexts: List[str] = ssm.list_teams_contexts(env_name=context.name)
            written_teams_contexts: List[str] = []
            for team in content["Teams"]:
                ssm.put_parameter(name=team["SsmParameterName"], obj=team)
                written_teams_contexts.append(team["SsmParameterName"])
            old_teams_contexts: List[str] = list(set(current_teams_contexts) - set(written_teams_contexts))
            _logger.debug("old_teams_contexts: %s", old_teams_contexts)
            ssm.delete_parameters(parameters=old_teams_contexts)
            del content["Teams"]
            ssm_parameter_name = context.ssm_parameter_name
        elif isinstance(context, FoundationContext):
            content = cast(Dict[str, Any], FoundationContext.Schema().dump(context))
            ssm_parameter_name = cast(str, context.ssm_parameter_name)
        else:
            raise ValueError("Unknown 'context' Type")

        ssm.put_parameter(name=ssm_parameter_name, obj=content)
        _logger.debug("Context written to SSM: %s", content)

    @staticmethod
    def load_context_from_ssm(env_name: str, type: Type[V]) -> V:
        if type is Context:
            context_parameter_name: str = f"/orbit/{env_name}/context"
            if ssm.does_parameter_exist(context_parameter_name):
                main = ssm.get_parameter(name=context_parameter_name)
            else:
                msg = f"SSM parameter {context_parameter_name} not found for env {env_name}"
                _logger.error(msg)
                raise Exception(msg)
            _logger.debug("Raw SSM: %s", main)
            teams_parameters = ssm.list_parameters(prefix=f"/orbit/{env_name}/teams/")
            _logger.debug("teams_parameters: %s", teams_parameters)
            teams = [ssm.get_parameter_if_exists(name=p) for p in teams_parameters if p.endswith("/context")]
            main["Teams"] = [t for t in teams if t]
            return cast(V, Context.Schema().load(data=main, many=False, partial=False, unknown="EXCLUDE"))
        elif type is FoundationContext:
            context_parameter_name = f"/orbit-f/{env_name}/context"
            main = ssm.get_parameter(name=context_parameter_name)
            return cast(V, FoundationContext.Schema().load(data=main, many=False, partial=False, unknown="EXCLUDE"))
        else:
            raise ValueError("Unknown 'context' Type")

    @staticmethod
    def fetch_toolkit_data(context: V) -> None:
        _logger.debug("Fetching Toolkit data...")
        if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
            raise ValueError("Unknown 'context' Type")

        top_level = "orbit" if isinstance(context, Context) else "orbit-f"

        resp_type = Dict[str, List[Dict[str, List[Dict[str, str]]]]]
        try:
            response: resp_type = boto3_client("cloudformation").describe_stacks(StackName=context.toolkit.stack_name)
            _logger.debug("%s stack found.", context.toolkit.stack_name)
        except botocore.exceptions.ClientError as ex:
            error: Dict[str, Any] = ex.response["Error"]
            if error["Code"] == "ValidationError" and f"{context.toolkit.stack_name} not found" in error["Message"]:
                _logger.debug("Toolkit stack not found.")
                return
            if (
                error["Code"] == "ValidationError"
                and f"{context.toolkit.stack_name} does not exist" in error["Message"]
            ):
                _logger.debug("Toolkit stack does not exist.")
                return
            raise
        if len(response["Stacks"]) < 1:
            _logger.debug("Toolkit stack not found.")
            return
        if "Outputs" not in response["Stacks"][0]:
            _logger.debug("Toolkit stack with empty outputs")
            return

        for output in response["Stacks"][0]["Outputs"]:
            if output["ExportName"] == f"{top_level}-{context.name}-deploy-id":
                _logger.debug("Export value: %s", output["OutputValue"])
                context.toolkit.deploy_id = output["OutputValue"]
            if output["ExportName"] == f"{top_level}-{context.name}-kms-arn":
                _logger.debug("Export value: %s", output["OutputValue"])
                context.toolkit.kms_arn = output["OutputValue"]
            if output["ExportName"] == f"{top_level}-{context.name}-admin-role":
                _logger.debug("Export value: %s", output["OutputValue"])
                context.toolkit.admin_role = output["OutputValue"]
            if output["ExportName"] == f"{top_level}-{context.name}-admin-role-arn":
                _logger.debug("Export value: %s", output["OutputValue"])
                context.toolkit.admin_role_arn = output["OutputValue"]

        if context.toolkit.deploy_id is None:
            raise RuntimeError(
                f"Stack {context.toolkit.stack_name} does not have the expected {top_level}-{context.name}-deploy-id output."
            )
        if context.toolkit.kms_arn is None:
            raise RuntimeError(
                f"Stack {context.toolkit.stack_name} does not have the expected {top_level}-{context.name}-kms-arn output."
            )
        if context.toolkit.admin_role is None:
            raise RuntimeError(
                f"Stack {context.toolkit.stack_name} does not have the expected {top_level}-{context.name}-admin-role output."
            )
        if context.toolkit.admin_role_arn is None:
            raise RuntimeError(
                f"Stack {context.toolkit.stack_name} does not have the expected {top_level}-{context.name}-admin-role-arn output."
            )
        context.toolkit.kms_alias = f"{top_level}-{context.name}-{context.toolkit.deploy_id}"
        context.toolkit.s3_bucket = (
            f"{top_level}-{context.name}-toolkit-{context.account_id}-{context.toolkit.deploy_id}"
        )
        context.cdk_toolkit.s3_bucket = (
            f"{top_level}-{context.name}-cdk-toolkit-{context.account_id}-{context.toolkit.deploy_id}"
        )

        if isinstance(context, Context):
            # Set the Helm repositories
            _logger.debug(f"context.helm_repository: s3://{context.toolkit.s3_bucket}/helm/repositories/env")
            context.helm_repository = f"s3://{context.toolkit.s3_bucket}/helm/repositories/env"

            _logger.debug("context.toolkit: %s", context.toolkit)

        _logger.debug("Toolkit data fetched successfully.")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from typing import Any, Dict, Optional

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def handler(event: Dict[str, Any], context: Optional[Dict[str, Any]]) -> Any:
    logger.info(f"Lambda metadata: {json.dumps(event)} (type = {type(event)})")

    namespace = event.get("Namespace")
    api = event.get("ExecutionInput", {}).get("Api")
    version = event.get("ExecutionInput", {}).get("Version", "v1")
    path = event.get("ExecutionInput", {}).get("Path")
    path_args = event.get("ExecutionInput", {}).get("PathArgs")
    method = event.get("ExecutionInput", {}).get("Method", "GET")
    query_parameters = event.get("ExecutionInput", {}).get("QueryParameters")
    request_body = event.get("ExecutionInput", {}).get("RequestBody")

    if namespace is None or path is None:
        raise ValueError("Parameters 'Namespace' and 'Path' are required")

    api = f"apis/{api}" if api else "api"
    base_path = f"/{api}/{version}/namespaces/{namespace}"
    path = path.format(**path_args) if path_args else path
    full_path = base_path + path
    logger.info(f"Reqeust: {method} {full_path}")
    return {"Path": full_path, "Method": method, "QueryParameters": query_parameters, "RequestBody": request_body}



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
from typing import Any, Dict, List, Optional

import boto3

logger = logging.getLogger()
logger.setLevel(logging.INFO)

REGION: str = os.environ["REGION"]


def get_nodegroups(cluster_name: str) -> List[Dict[str, Dict[str, str]]]:
    return_response: List[Dict[str, Dict[str, str]]] = []
    eks_client = boto3.client("eks")
    try:
        nodegroups_response = eks_client.list_nodegroups(clusterName=cluster_name)
        for nodegroup_name in nodegroups_response["nodegroups"]:
            nodegroup_details = eks_client.describe_nodegroup(clusterName=cluster_name, nodegroupName=nodegroup_name)
            if "nodegroup" in nodegroup_details:
                nodegroup = nodegroup_details["nodegroup"]
                nodegroup_dict = {
                    "nodegroup_name": nodegroup_name,
                    "instance_types": nodegroup["instanceTypes"],
                    "scaling_config": json.dumps(nodegroup["scalingConfig"]).encode("utf-8"),
                    "status": nodegroup["status"],
                    "capacity_type": nodegroup["capacityType"],
                }
                # Get Disk size
                if "diskSize" in nodegroup:
                    nodegroup_dict["disk_size"] = nodegroup["diskSize"]
                # Get Launch Template details
                if "launchTemplate" in nodegroup:
                    nodegroup_dict["launch_template"] = nodegroup["launchTemplate"]

                return_response.append(nodegroup_dict)
    except Exception as ekse:
        logger.error("Error describing cluster %s nodegroups: %s", cluster_name, ekse)
        raise ekse
    return return_response


def handler(event: Dict[str, Any], context: Optional[Dict[str, Any]]) -> List[Dict[str, Dict[str, str]]]:
    cluster_name = event["cluster_name"]
    nodegroups: List[Dict[str, Dict[str, str]]] = get_nodegroups(cluster_name=cluster_name)
    return nodegroups



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# flake8: noqa: F811

import json
import logging
import os
import re
from typing import Any, ClassVar, Dict, Generic, List, Optional, Set, Type, TypeVar, Union, cast

import jsonpath_ng as jsonpath_ng
import yaml
from dataclasses import field
from marshmallow import EXCLUDE, Schema
from marshmallow_dataclass import dataclass
from yamlinclude import YamlIncludeConstructor

import aws_orbit
from aws_orbit import utils
from aws_orbit.models.common import BaseSchema
from aws_orbit.services import ssm
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)

SSM_CONTEXT: Dict[str, str] = {}


@dataclass(base_schema=BaseSchema, frozen=True)
class PluginManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    plugin_id: str
    module: str
    parameters: Dict[str, Any] = field(default_factory=dict)
    path: Optional[str] = None


@dataclass(base_schema=BaseSchema, frozen=True)
class TeamManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    policies: List[str] = field(default_factory=list)
    grant_sudo: bool = False
    fargate: bool = True
    k8_admin: bool = False
    jupyterhub_inbound_ranges: List[str] = field(default_factory=lambda: ["0.0.0.0/0"])
    image: Optional[str] = None
    plugins: List[PluginManifest] = field(default_factory=list)
    efs_life_cycle: Optional[str] = None
    authentication_groups: Optional[List[str]] = None


@dataclass(base_schema=BaseSchema, frozen=True)
class ImageManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    repository: Optional[str]
    version: Optional[str] = "latest" if aws_orbit.__version__.endswith(".dev0") else aws_orbit.__version__
    path: Optional[str] = None

    def get_source(self, account_id: str, region: str) -> str:
        external_ecr_match = re.compile(r"^[0-9]{12}\.dkr\.ecr\..+\.amazonaws.com/")
        public_ecr_match = re.compile(r"^public.ecr.aws/.+/")
        repository = cast(str, self.repository)

        if self.path:
            return "code"
        elif repository.startswith(f"{account_id}.dkr.ecr.{region}.amazonaws.com"):
            return "ecr-internal"
        elif external_ecr_match.match(repository):
            return "ecr-external"
        elif public_ecr_match.match(repository):
            return "ecr-public"
        else:
            return "public"


@dataclass(base_schema=BaseSchema, frozen=True)
class ManagedNodeGroupManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    instance_type: str = "m5.4xlarge"
    local_storage_size: int = 128
    nodes_num_desired: int = 2
    nodes_num_max: int = 3
    nodes_num_min: int = 1
    labels: Dict[str, str] = field(default_factory=dict)
    enable_virtual_gpu: bool = False
    availability_zones: Optional[List[str]] = None
    efa_enabled: Optional[bool] = None


@dataclass(base_schema=BaseSchema, frozen=True)
class CodeBuildImageManifest(ImageManifest):
    repository: Optional[str] = "public.ecr.aws/v3o4w1g6/aws-orbit-workbench/code-build-base"
    version: Optional[str] = "1.1.0"


@dataclass(base_schema=BaseSchema, frozen=True)
class JupyterUserImageManifest(ImageManifest):
    repository: Optional[str] = "public.ecr.aws/v3o4w1g6/aws-orbit-workbench/jupyter-user"


@dataclass(base_schema=BaseSchema, frozen=True)
class OrbitControllerImageManifest(ImageManifest):
    repository: Optional[str] = "public.ecr.aws/v3o4w1g6/aws-orbit-workbench/orbit-controller"


@dataclass(base_schema=BaseSchema, frozen=True)
class K8sUtilitiesImageManifest(ImageManifest):
    repository: Optional[str] = "public.ecr.aws/v3o4w1g6/aws-orbit-workbench/k8s-utilities"
    version: Optional[str] = "1.0.0"


@dataclass(base_schema=BaseSchema, frozen=True)
class UtilityDataImageManifest(ImageManifest):
    repository: Optional[str] = "public.ecr.aws/v3o4w1g6/aws-orbit-workbench/utility-data"
    version: Optional[str] = "1.0.0"


@dataclass(base_schema=BaseSchema, frozen=True)
class FoundationImagesManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    code_build: CodeBuildImageManifest = CodeBuildImageManifest()
    names: List[str] = field(
        metadata=dict(load_only=True),
        default_factory=lambda: [
            "code_build",
        ],
    )


@dataclass(base_schema=BaseSchema, frozen=True)
class ImagesManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    code_build: ImageManifest = CodeBuildImageManifest()
    jupyter_user: ImageManifest = JupyterUserImageManifest()
    orbit_controller: ImageManifest = OrbitControllerImageManifest()
    k8s_utilities: ImageManifest = K8sUtilitiesImageManifest()
    utility_data: ImageManifest = UtilityDataImageManifest()
    names: List[str] = field(
        metadata=dict(load_only=True),
        default_factory=lambda: [
            "code_build",
            "jupyter_user",
            "orbit_controller",
            "k8s_utilities",
            "utility_data",
        ],
    )


@dataclass(base_schema=BaseSchema, frozen=True)
class FrontendNetworkingManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    load_balancers_subnets: List[str] = field(default_factory=list)
    ssl_cert_arn: Optional[str] = None
    custom_domain_name: Optional[str] = None


@dataclass(base_schema=BaseSchema, frozen=True)
class DataNetworkingManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    internet_accessible: bool = True
    nodes_subnets: List[str] = field(default_factory=list)


@dataclass(base_schema=BaseSchema, frozen=True)
class NetworkingManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    vpc_id: Optional[str] = None
    public_subnets: List[str] = field(default_factory=list)
    private_subnets: List[str] = field(default_factory=list)
    isolated_subnets: List[str] = field(default_factory=list)
    max_availability_zones: Optional[int] = None
    frontend: FrontendNetworkingManifest = FrontendNetworkingManifest()
    data: DataNetworkingManifest = DataNetworkingManifest()


@dataclass(base_schema=BaseSchema, frozen=True)
class FoundationManifest:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    role_prefix: Optional[str] = None
    codeartifact_domain: Optional[str] = None
    codeartifact_repository: Optional[str] = None
    images: FoundationImagesManifest = FoundationImagesManifest()
    policies: Optional[List[str]] = cast(List[str], field(default_factory=list))
    ssm_parameter_name: Optional[str] = None
    networking: NetworkingManifest = NetworkingManifest()


@dataclass(base_schema=BaseSchema, frozen=True)
class Manifest:
    Schema: ClassVar[Type[Schema]] = Schema
    name: str
    role_prefix: Optional[str] = None
    user_pool_id: Optional[str] = None
    scratch_bucket_arn: Optional[str] = None
    eks_system_masters_roles: Optional[List[str]] = cast(List[str], field(default_factory=list))
    codeartifact_domain: Optional[str] = None
    codeartifact_repository: Optional[str] = None
    cognito_external_provider: Optional[str] = None
    cognito_external_provider_label: Optional[str] = None
    networking: NetworkingManifest = NetworkingManifest()
    teams: List[TeamManifest] = field(default_factory=list)
    images: ImagesManifest = ImagesManifest()
    shared_efs_fs_id: Optional[str] = None
    shared_efs_sg_id: Optional[str] = None
    managed_nodegroups: List[ManagedNodeGroupManifest] = field(default_factory=list)
    policies: Optional[List[str]] = cast(List[str], field(default_factory=list))
    ssm_parameter_name: Optional[str] = None
    install_ssm_agent: Optional[bool] = False
    install_image_replicator: Optional[bool] = False

    def get_team_by_name(self, name: str) -> Optional[TeamManifest]:
        for t in self.teams:
            if t.name == name:
                return t
        return None


def _add_ssm_param_injector(tag: str = "!SSM") -> Set[str]:
    """
    Load a yaml configuration file and resolve any SSM parameters
    The SSM parameters must have !SSM before them and be in this format
    to be parsed: ${SSM_PARAMETER_PATH::JSONPATH}.
    E.g.:
    database:
        host: !SSM ${/orbit-f/dev-env/resources::/UserAccessPolicy}
        port: !SSM ${/orbit-f/dev-env/resources::/PublicSubnet/*}
    """
    # pattern for global vars: look for ${word}
    pattern = re.compile(".*?\${([^}]+::[^}]*)}.*?")  # noqa: W605
    loader = yaml.SafeLoader

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !SSM somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, pattern, None)  # type: ignore

    ssm_parameters = set()

    def constructor_ssm_parameter(loader, node) -> Any:  # type: ignore
        """
        Extracts the environment variable from the node's value
        :param yaml.Loader loader: the yaml loader
        :param node: the current node in the yaml
        :return: the parsed string that contains the value of the environment
        variable
        """
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if match:
            full_value = value
            for g in match:
                _logger.debug(f"match: {g}")
                (ssm_param_name, jsonpath) = g.split("::")
                if "${" in ssm_param_name:
                    ssm_param_name = ssm_param_name.replace("$", "").format(os.environ)
                _logger.debug(f"found injected parameter {(ssm_param_name, jsonpath)}")
                if ssm_param_name not in SSM_CONTEXT:
                    ssm = boto3_client("ssm")
                    try:
                        SSM_CONTEXT[ssm_param_name] = json.loads(
                            ssm.get_parameter(Name=ssm_param_name)["Parameter"]["Value"]
                        )
                        ssm_parameters.add(ssm_param_name)
                    except Exception as e:
                        _logger.error(f"Error resolving injected parameter {g}: {e}")

                json_expr = jsonpath_ng.parse(jsonpath)
                json_data = SSM_CONTEXT[ssm_param_name]
                json_match = json_expr.find(json_data)

                if len(json_match) > 1:
                    raise Exception(f"Injected parameter {g} is ambiguous")
                elif len(json_match) == 0:
                    raise Exception(f"Injected parameter {jsonpath} not found in SSM {ssm_param_name}")

                param_value: str = json_match[0].value
                _logger.debug(f"injected SSM parameter {g} resolved to {param_value}")
                return param_value
            return full_value
        return value

    loader.add_constructor(tag, constructor_ssm_parameter)  # type: ignore
    return ssm_parameters


def _add_env_var_injector(tag: str = "!ENV") -> None:
    """
    Load a yaml configuration file and resolve any environment variables
    The environment variables must have !ENV before them and be in this format
    to be parsed: ${VAR_NAME}.
    E.g.:
    database:
        host: !ENV ${HOST}
        port: !ENV ${PORT}
    app:
        log_path: !ENV '/var/${LOG_PATH}'
        something_else: !ENV '${AWESOME_ENV_VAR}/var/${A_SECOND_AWESOME_VAR}'
    """
    # pattern for global vars: look for ${word}
    pattern = re.compile(".*?\${([^}]+::[^}]*)}.*?")  # noqa: W605
    loader = yaml.SafeLoader

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !ENV somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, pattern, None)  # type: ignore

    def constructor_env_variables(loader, node) -> Any:  # type: ignore
        """
        Extracts the environment variable from the node's value
        :param yaml.Loader loader: the yaml loader
        :param node: the current node in the yaml
        :return: the parsed string that contains the value of the environment
        variable
        """
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if match:
            full_value = value
            for g in match:
                (env_var, default_val) = g.split("::")
                value = os.environ.get(env_var, default_val)
                full_value = full_value.replace(f"${{{g}}}", value)
                if not full_value:
                    full_value = None
                _logger.debug(f"injected ENV parameter {env_var} resolved to {value}")
            return full_value
        return value

    loader.add_constructor(tag, constructor_env_variables)  # type: ignore


T = TypeVar("T")


class ManifestSerDe(Generic[T]):
    @staticmethod
    def load_manifest_from_file(filename: str, type: Type[T]) -> T:
        _logger.debug("Loading manifest file (%s)", filename)
        filepath = os.path.abspath(filename)
        _logger.debug("filepath: %s", filepath)
        filedir: str = os.path.dirname(filepath)
        utils.print_dir(dir=filedir)
        YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.SafeLoader, base_dir=filedir)
        _add_ssm_param_injector()
        _add_env_var_injector()
        with open(filepath, "r") as f:
            raw: Dict[str, Any] = cast(Dict[str, Any], yaml.safe_load(f))
        _logger.debug("raw: %s", raw)
        if type is Manifest:
            raw["SsmParameterName"] = f"/orbit/{raw['Name']}/manifest"
            manifest: T = cast(T, Manifest.Schema().load(data=raw, many=False, partial=False, unknown=EXCLUDE))
        elif type is FoundationManifest:
            raw["SsmParameterName"] = f"/orbit-f/{raw['Name']}/manifest"
            manifest = cast(T, FoundationManifest.Schema().load(data=raw, many=False, partial=False, unknown=EXCLUDE))
        else:
            raise ValueError("Unknown 'manifest' Type")
        ManifestSerDe.dump_manifest_to_ssm(manifest=manifest)
        return manifest

    @staticmethod
    def dump_manifest_to_file(manifest: T, filepath: str) -> None:
        _logger.debug("Dumping manifest file (%s)", filepath)
        if isinstance(manifest, Manifest):
            content: Dict[str, Any] = cast(Dict[str, Any], Manifest.Schema().dump(manifest))
        elif isinstance(manifest, FoundationManifest):
            content = cast(Dict[str, Any], FoundationManifest.Schema().dump(manifest))
        else:
            raise ValueError("Unknown 'manifest' Type")
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, "w") as f:
            yaml.dump(content, f, sort_keys=False)

    @staticmethod
    def dump_manifest_to_ssm(manifest: T) -> None:
        _logger.debug("Writing manifest to SSM parameter.")

        if isinstance(manifest, Manifest):
            _logger.debug("Teams: %s", [t.name for t in manifest.teams])
            content: Dict[str, Any] = cast(Dict[str, Any], Manifest.Schema().dump(manifest))
            ssm.cleanup_manifest(env_name=manifest.name)
            if "Teams" in content:
                for team in content["Teams"]:
                    team_parameter_name: str = f"/orbit/{manifest.name}/teams/{team['Name']}/manifest"
                    ssm.put_parameter(name=team_parameter_name, obj=team)
                del content["Teams"]
            manifest_parameter_name = manifest.ssm_parameter_name
        elif isinstance(manifest, FoundationManifest):
            content = cast(Dict[str, Any], FoundationManifest.Schema().dump(manifest))
            ssm.cleanup_manifest(env_name=manifest.name, top_level="orbit-f")
            manifest_parameter_name = manifest.ssm_parameter_name
        else:
            raise ValueError("Unknown 'manifest' Type")

        ssm.put_parameter(name=cast(str, manifest_parameter_name), obj=content)

    @staticmethod
    def load_manifest_from_ssm(env_name: str, type: Type[T]) -> Optional[T]:
        if type is Manifest:
            context_parameter_name = f"/orbit/{env_name}/manifest"
            main = ssm.get_parameter_if_exists(name=context_parameter_name)
            if main is None:
                return None
            teams_parameters = ssm.list_parameters(prefix=f"/orbit/{env_name}/teams/")
            _logger.debug("teams_parameters (/orbit/%s/teams/): %s", env_name, teams_parameters)
            teams = [ssm.get_parameter(name=p) for p in teams_parameters if p.endswith("/manifest")]
            main["Teams"] = teams
            return cast(T, Manifest.Schema().load(data=main, many=False, partial=False, unknown=EXCLUDE))
        elif type is FoundationManifest:
            context_parameter_name = f"/orbit-f/{env_name}/manifest"
            main = ssm.get_parameter_if_exists(name=context_parameter_name)
            if main is None:
                return None
            return cast(T, FoundationManifest.Schema().load(data=main, many=False, partial=False, unknown=EXCLUDE))
        else:
            raise ValueError("Unknown 'manifest' Type")


def manifest_validations(manifest: Manifest) -> None:
    for managed_nodegroup in manifest.managed_nodegroups:
        _logger.debug(f"Validating managed node group {managed_nodegroup} ")
        if managed_nodegroup.nodes_num_desired < 1:
            raise ValueError(f"{managed_nodegroup.name}  number of desired nodes should be greater than 0")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from typing import List, TypeVar

from aws_orbit import sh
from aws_orbit.models.context import Context, FoundationContext
from aws_orbit.services import cfn

_logger: logging.Logger = logging.getLogger(__name__)

T = TypeVar("T")


def get_app_argument(app_filename: str, args: List[str]) -> str:
    args_str: str = " ".join(args)
    return f'--app "python {app_filename} {args_str}"'


def get_output_argument(context: T, stack_name: str) -> str:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    path: str = os.path.join(".orbit.out", context.name, "cdk", stack_name)
    return f"--output {path}"


def deploy(context: T, stack_name: str, app_filename: str, args: List[str]) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    cmd: str = (
        "cdk deploy --require-approval never --progress events "
        f"--toolkit-stack-name {context.cdk_toolkit.stack_name} "
        f"{get_app_argument(app_filename, args)} "
        f"{get_output_argument(context, stack_name)}"
    )
    sh.run(cmd=cmd)


def destroy(context: T, stack_name: str, app_filename: str, args: List[str]) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    if cfn.does_stack_exist(stack_name=stack_name) is False:
        _logger.debug("Skipping CDK destroy for %s, because the stack was not found.", stack_name)
        return
    cmd: str = (
        "cdk destroy --force "
        f"--toolkit-stack-name {context.cdk_toolkit.stack_name} "
        f"{get_app_argument(app_filename, args)} "
        f"{get_output_argument(context, stack_name)}"
    )
    sh.run(cmd=cmd)


def deploy_toolkit(context: T) -> None:
    if not (isinstance(context, Context) or isinstance(context, FoundationContext)):
        raise ValueError("Unknown 'context' Type")
    cmd: str = (
        f"cdk bootstrap --toolkit-bucket-name {context.cdk_toolkit.s3_bucket} "
        f"--toolkit-stack-name {context.cdk_toolkit.stack_name} "
        f"{get_output_argument(context, context.cdk_toolkit.stack_name)} "
        f"aws://{context.account_id}/{context.region}"
    )
    sh.run(cmd=cmd)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List

from aws_orbit.plugins import PLUGINS_REGISTRIES

if TYPE_CHECKING:
    from aws_orbit.models.context import Context, TeamContext

_logger: logging.Logger = logging.getLogger(__name__)


def deploy(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]:
    PLUGINS_REGISTRIES.add_hook(hook_name="deploy_hook", func=func)
    return func


def destroy(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]:
    PLUGINS_REGISTRIES.add_hook(hook_name="destroy_hook", func=func)
    return func


def dockerfile_injection(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], List[str]]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], List[str]]:
    PLUGINS_REGISTRIES.add_hook(hook_name="dockerfile_injection_hook", func=func)
    return func


def bootstrap_injection(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], str]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], str]:
    PLUGINS_REGISTRIES.add_hook(hook_name="bootstrap_injection_hook", func=func)
    return func


def pre(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]:
    PLUGINS_REGISTRIES.add_hook(hook_name="pre_hook", func=func)
    return func


def post(
    func: Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]
) -> Callable[[str, "Context", "TeamContext", Dict[str, Any]], None]:
    PLUGINS_REGISTRIES.add_hook(hook_name="post_hook", func=func)
    return func



import logging
import time
from typing import Any, Dict, List, cast

from kubernetes import config
from kubernetes.client import AppsV1Api, CoreV1Api, NetworkingV1beta1Api, NetworkingV1beta1IngressList, V1Service

_logger: logging.Logger = logging.getLogger(__name__)


def get_service_hostname(name: str, k8s_context: str, namespace: str = "default") -> str:
    config.load_kube_config(context=k8s_context)
    v1 = CoreV1Api()
    while True:
        resp = cast(V1Service, v1.read_namespaced_service_status(name=name, namespace=namespace))
        status: Dict[str, Any] = resp.status.to_dict()
        if "load_balancer" in status:
            if "ingress" in status["load_balancer"]:
                if status["load_balancer"]["ingress"]:
                    if "hostname" in status["load_balancer"]["ingress"][0]:
                        break
        time.sleep(10)
    return str(status["load_balancer"]["ingress"][0]["hostname"])


def get_service_addresses(name: str, k8s_context: str, namespace: str = "orbit-system") -> List[Dict[str, Any]]:
    _logger.debug("Retrieving Endpoints for Service %s in Namespace %s", name, namespace)
    config.load_kube_config(context=k8s_context)
    v1 = CoreV1Api()
    result = []
    for attempt in range(10):
        time.sleep(60)
        resp = v1.read_namespaced_endpoints(name=name, namespace=namespace)
        endpoints = resp.to_dict()
        subsets = endpoints.get("subsets")
        if subsets and len(subsets) > 0:
            _logger.debug("Endpoint Subsets: %s", subsets)

            addresses = None
            for subset in subsets:
                addresses = subset.get("addresses")
                if addresses and len(addresses) > 0:
                    result.extend(addresses)

            if addresses and len(addresses) > 0:
                return cast(List[Dict[str, Any]], addresses)

        _logger.debug(
            "No Addresses found for Service %s in Namespace %s, sleeping for 1 minute (attempt: %s)",
            name,
            namespace,
            attempt,
        )
    else:
        return []


def get_ingress_dns(name: str, k8s_context: str, namespace: str = "default") -> str:
    config.load_kube_config(context=k8s_context)

    network = NetworkingV1beta1Api()
    for attempt in range(15):
        time.sleep(60)
        try:
            resp: NetworkingV1beta1IngressList = network.list_namespaced_ingress(namespace=namespace)
            _logger.debug(resp)
            for i in resp.items:
                item: Dict[str, Any] = i.to_dict()
                if item["metadata"]["name"] == name:
                    if "load_balancer" in item["status"]:
                        if "ingress" in item["status"]["load_balancer"]:
                            if item["status"]["load_balancer"]["ingress"]:
                                if "hostname" in item["status"]["load_balancer"]["ingress"][0]:
                                    return str(item["status"]["load_balancer"]["ingress"][0]["hostname"])
                                else:
                                    _logger.debug("hostname is not defined ")
                            else:
                                _logger.debug("ingress[] is empty")
                        else:
                            _logger.debug("ingress not in load_balancer")
                    else:
                        _logger.debug("no load_balancer in status")
            else:
                _logger.debug(f"Cannot find Ingress {name}.{namespace}")
        except Exception:
            _logger.exception(f"Error finding Ingress {name}.{namespace}")

        _logger.info("Waiting for for Ingress %s.%s, sleeping for 1 minute (attempt: %s)", name, namespace, attempt)
    else:
        raise Exception(f"Timeout while waiting for Ingress {name}.{namespace}")


def get_resource_status(name: str, k8s_context: str, type: str, namespace: str = "orbit-system") -> Dict[str, Any]:
    _logger.debug("Retrieving Status for %s %s in Namespace %s", type, name, namespace)
    config.load_kube_config(context=k8s_context)
    apps = AppsV1Api()
    if type.lower() == "statefulset":
        api = apps.read_namespaced_stateful_set_status
    elif type.lower() == "deployment":
        api = apps.read_namespaced_deployment_status
    else:
        raise Exception("Unknown resource type")

    resp = api(name=name, namespace=namespace)
    resource = resp.to_dict()
    return cast(Dict[str, Any], resource.get("status"))


if __name__ == "__main__":
    k8s_context = config.load_kube_config()
    r = get_ingress_dns(name="istio-ingress", k8s_context=k8s_context, namespace="istio-system")
    _logger.debug(r)



import base64
import logging
from typing import cast

from aws_orbit.models.context import Context
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def encrypt(context: Context, plaintext: str) -> str:
    client = boto3_client("kms")

    _logger.debug("Encrypting data")
    response = client.encrypt(KeyId=context.toolkit.kms_arn, Plaintext=plaintext.encode("utf-8"))
    return base64.b64encode(response.get("CiphertextBlob")).decode("utf-8")


def decrypt(context: Context, ciphertext: str) -> str:
    client = boto3_client("kms")

    _logger.debug("Decrypting data")
    response = client.decrypt(KeyId=context.toolkit.kms_arn, CiphertextBlob=base64.b64decode(ciphertext))
    return cast(str, response["Plaintext"].decode("utf-8"))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from marshmallow import EXCLUDE, Schema, fields


class BaseSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    @staticmethod
    def _from_snake_to_camel(name: str) -> str:
        return "".join(map(str.title, name.split("_")))

    def on_bind_field(self, field_name: str, field_obj: fields.Field) -> None:
        field_obj.data_key = BaseSchema._from_snake_to_camel(field_obj.data_key or field_name)



import logging
import shlex
import subprocess
from typing import Iterable, Optional

from aws_orbit.exceptions import FailedShellCommand

_logger: logging.Logger = logging.getLogger(__name__)


def _clean_up_stdout_line(line: bytes) -> str:
    line_str = line.decode("utf-8")
    return line_str[:-1] if line_str.endswith("\n") else line_str


def _run_iterating(cmd: str, cwd: Optional[str] = None) -> Iterable[str]:
    p = subprocess.Popen(shlex.split(cmd), cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    if p.stdout is None:
        return []
    while p.poll() is None:
        yield _clean_up_stdout_line(line=p.stdout.readline())
    if p.returncode != 0:
        raise FailedShellCommand(f"Exit code: {p.returncode}")


def run(cmd: str, cwd: Optional[str] = None, hide_cmd: bool = False) -> None:
    if hide_cmd is False:
        _logger.debug(f"+ {cmd}")
    for line in _run_iterating(cmd=cmd, cwd=cwd):
        _logger.debug(line)


def run_iterating(cmd: str, cwd: Optional[str] = None, hide_cmd: bool = False) -> Iterable[str]:
    if hide_cmd is False:
        _logger.debug(f"+ {cmd}")
    for line in _run_iterating(cmd=cmd, cwd=cwd):
        _logger.debug(line)
        yield line
